# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2013-2016 Colin Duquesnoy and others (see pyqode/AUTHORS.rst)
# Copyright (c) 2016- Spyder Project Contributors (see AUTHORS.txt)
#
# Distributed under the terms of the MIT License
# (see NOTICE.txt in the Spyder root directory for details)
# -----------------------------------------------------------------------------

"""
This module contains the panel API.

Adapted from pyqode/core/api/panel.py of the
`PyQode project <https://github.com/pyQode/pyQode>`_.
Original file:
<https://github.com/pyQode/pyqode.core/blob/master/pyqode/core/api/panel.py>
"""

# Standard library imports
from math import ceil
import logging

# Third party imports
from qtpy.QtWidgets import QWidget, QApplication
from qtpy.QtGui import QBrush, QColor, QPen, QPainter
from qtpy.QtCore import Qt, QRect

# Local imports
from spyder.plugins.editor.api.editorextension import EditorExtension
from spyder.plugins.outlineexplorer.api import is_cell_header

logger = logging.getLogger(__name__)


class Panel(QWidget, EditorExtension):
    """
    Base class for editor panels.

    A panel is a editor extension and a QWidget.

    .. note:: Use enabled to disable panel actions and setVisible to change the
        visibility of the panel.
    """
    class Position(object):
        """Enumerates the possible panel positions"""
        # Top margin
        TOP = 0
        # Left margin
        LEFT = 1
        # Right margin
        RIGHT = 2
        # Bottom margin
        BOTTOM = 3
        # Floating panel
        FLOATING = 4

        @classmethod
        def iterable(cls):
            """ Returns possible positions as an iterable (list) """
            return [cls.TOP, cls.LEFT, cls.RIGHT, cls.BOTTOM]

    @property
    def scrollable(self):
        """
        A scrollable panel will follow the editor's scroll-bars.

        Left and right panels follow the vertical scrollbar. Top and bottom
        panels follow the horizontal scrollbar.

        :type: bool
        """
        return self._scrollable

    @scrollable.setter
    def scrollable(self, value):
        self._scrollable = value

    def __init__(self, dynamic=False):
        EditorExtension.__init__(self)
        QWidget.__init__(self)
        # Specifies whether the panel is dynamic. A dynamic panel is a panel
        # that will be shown/hidden depending on the context.
        # Dynamic panel should not appear in any GUI menu
        self.dynamic = dynamic
        # Panel order into the zone it is installed to. This value is
        # automatically set when installing the panel but it can be changed
        # later (negative values can also be used).
        self.order_in_zone = -1
        self._scrollable = False
        self._background_brush = None
        self.linecell_color = QColor(Qt.darkGray)
        self._foreground_pen = None
        # Position in the editor (top, left, right, bottom)
        self.position = -1

    def on_install(self, editor):
        """
        Extends :meth:`spyder.api.EditorExtension.on_install` method to set the
        editor instance as the parent widget.

        .. warning:: Don't forget to call **super** if you override this
            method!

        :param editor: editor instance
        :type editor: spyder.plugins.editor.widgets.codeeditor.CodeEditor
        """
        EditorExtension.on_install(self, editor)
        self.setParent(editor)
        self.setPalette(QApplication.instance().palette())
        self.setFont(QApplication.instance().font())
        self.editor.panels.refresh()
        self._background_brush = QBrush(QColor(
            self.palette().window().color()))
        self._foreground_pen = QPen(QColor(
            self.palette().windowText().color()))

        if self.position == self.Position.FLOATING:
            self.setAttribute(Qt.WA_TransparentForMouseEvents)

    def paintEvent(self, event):
        """
        Fill the panel background using QPalette.

        Notes
        -----
        Please remember to extend this method in the child class to
        paint the panel's desired information.
        """
        if self.isVisible() and self.position != self.Position.FLOATING:
            # fill background
            self._background_brush = QBrush(QColor(
                self.editor.sideareas_color))
            self._foreground_pen = QPen(QColor(
                self.palette().windowText().color()))
            painter = QPainter(self)
            painter.fillRect(event.rect(), self._background_brush)
        else:
            logger.debug(f'paintEvent method must be defined in {self}')

    def paint_cell(self, painter):
        """Paint cell dividers in the visible region if needed."""
        for top_position, line_number, block in self.editor.visible_blocks:
            if (
                is_cell_header(block)
                and (self.position in [self.Position.LEFT, self.Position.RIGHT])
            ):
                pen = painter.pen()
                pen.setStyle(Qt.SolidLine)
                pen.setBrush(self.linecell_color)
                painter.setPen(pen)
                painter.drawLine(0, top_position, self.width(),
                                 top_position)


    def sizeHint(self):
        """
        Return the widget size hint, overriding the Qt method.

        Notes
        -----
        * This size hint will define the QSize of the panel, i.e. it is
          where its width and height are defined.
        * If the size of your panel depends on displayed text, please
          use the LineNumberArea one as reference on how to implement
          this method.
        * If the size is not dependent on displayed text, please use
          the debugger panel as reference.
        * If your panel is in a floating position, please use the
          IndentationGuide one as reference.
        """
        if self.position != self.Position.FLOATING:
            raise NotImplementedError(
                f'sizeHint method must be implemented in {self}')

    def setVisible(self, visible):
        """
        Shows/Hides the panel.

        Automatically call PanelsManager.refresh_panels.

        :param visible: Visible state
        """
        logger.debug('%s visibility changed', self.name)
        super(Panel, self).setVisible(visible)
        if self.editor:
            self.editor.panels.refresh()

    def geometry(self):
        """Return geometry dimensions for floating Panels.

        Note: If None is returned It'll use editor contentsRect dimensions.

        returns: x0, y0, height width.
        """
        return 0, 0, None, None

    def set_geometry(self, crect):
        """Set geometry for floating panels.

        Normally you don't need to override this method, you should override
        `geometry` instead.
        """
        x0, y0, width, height = self.geometry()

        if width is None:
            width = crect.width()
        if height is None:
            height = crect.height()

        # Calculate editor coordinates with their offsets
        offset = self.editor.contentOffset()
        x = self.editor.blockBoundingGeometry(self.editor.firstVisibleBlock())\
            .translated(offset.x(), offset.y()).left() \
            + self.editor.document().documentMargin() \
            + self.editor.panels.margin_size(Panel.Position.LEFT)
        y = crect.top() + self.editor.panels.margin_size(Panel.Position.TOP)

        self.setGeometry(QRect(ceil(x+x0), ceil(y+y0),
                               ceil(width), ceil(height)))



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for cli_options.py
"""
import pytest
from spyder.app import cli_options


@pytest.mark.order(1)
def test_get_options():
    getopt = cli_options.get_options

    options, args = getopt([])
    assert not options.new_instance
    assert not options.reset_to_defaults
    assert not options.reset_config_files
    assert not options.optimize
    assert not options.paths
    assert options.working_directory is None
    assert not options.hide_console
    assert not options.show_console
    assert not options.multithreaded
    assert not options.profile
    assert options.window_title is None
    assert options.project is None
    assert options.opengl_implementation is None
    assert options.files == []
    assert args == []

    options, args = getopt(['--new-instance'])
    assert options.new_instance

    options, args = getopt(['--defaults', '--reset'])
    assert options.reset_to_defaults
    assert options.reset_config_files

    options, args = getopt(['--optimize', '--workdir', 'test dir'])
    assert options.optimize
    assert options.working_directory == 'test dir'

    options, args = getopt('--window-title MyWindow'.split())
    assert options.window_title == 'MyWindow'

    options, args = getopt('-p myproject test_file.py another_file.py'.split())
    assert options.project == 'myproject'
    assert options.files == ['test_file.py', 'another_file.py']
    assert args == ['test_file.py', 'another_file.py']

    with pytest.raises(SystemExit):
        options, args = getopt(['--version'])

    # Requires string.
    with pytest.raises(SystemExit):
        options, args = getopt(['-w'])

    # Requires string.
    with pytest.raises(SystemExit):
        options, args = getopt(['-p'])

    options, args = getopt('--opengl software'.split())
    assert options.opengl_implementation == 'software'


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Mixin to connect a plugin to the IPython console.
"""

from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.plugins import Plugins


class ShellConnectMixin:
    """
    Mixin to connect any widget or object to the shell widgets in the IPython
    console.
    """

    # ---- Connection to the IPython console
    # -------------------------------------------------------------------------
    def register_ipythonconsole(self, ipyconsole):
        """Register signals from the console."""
        ipyconsole.sig_shellwidget_changed.connect(self.set_shellwidget)
        ipyconsole.sig_shellwidget_created.connect(self.add_shellwidget)
        ipyconsole.sig_shellwidget_deleted.connect(self.remove_shellwidget)
        ipyconsole.sig_shellwidget_errored.connect(
            self.add_errored_shellwidget)

    def unregister_ipythonconsole(self, ipyconsole):
        """Unregister signals from the console."""
        ipyconsole.sig_shellwidget_changed.disconnect(self.set_shellwidget)
        ipyconsole.sig_shellwidget_created.disconnect(self.add_shellwidget)
        ipyconsole.sig_shellwidget_deleted.disconnect(self.remove_shellwidget)
        ipyconsole.sig_shellwidget_errored.disconnect(
            self.add_errored_shellwidget)

    # ---- Public API
    # -------------------------------------------------------------------------
    def set_shellwidget(self, shellwidget):
        """Update the current shellwidget."""
        raise NotImplementedError

    def add_shellwidget(self, shellwidget):
        """Add a new shellwidget to be registered."""
        raise NotImplementedError

    def remove_shellwidget(self, shellwidget):
        """Remove a registered shellwidget."""
        raise NotImplementedError

    def add_errored_shellwidget(self, shellwidget):
        """Register a new shellwidget whose kernel failed to start."""
        raise NotImplementedError


class ShellConnectPluginMixin(ShellConnectMixin):
    """
    Mixin to connect a plugin composed of stacked widgets to the shell widgets
    in the IPython console.

    It is assumed that self.get_widget() returns an instance of
    ShellConnectMainWidget.
    """

    # ---- Connection to the IPython console
    # -------------------------------------------------------------------------
    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipython_console_available(self):
        """Connect to the IPython console."""
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        self.register_ipythonconsole(ipyconsole)

    @on_plugin_teardown(plugin=Plugins.IPythonConsole)
    def on_ipython_console_teardown(self):
        """Disconnect from the IPython console."""
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        self.unregister_ipythonconsole(ipyconsole)

    # ---- Public API
    # -------------------------------------------------------------------------
    def set_shellwidget(self, shellwidget):
        """
        Update the current shellwidget.

        Parameters
        ----------
        shellwidget: spyder.plugins.ipyconsole.widgets.shell.ShellWidget
            The shell widget.
        """
        self.get_widget().set_shellwidget(shellwidget)

    def add_shellwidget(self, shellwidget):
        """
        Add a new shellwidget to be registered.

        This function registers a new widget to display content that
        comes from shellwidget.

        Parameters
        ----------
        shellwidget: spyder.plugins.ipyconsole.widgets.shell.ShellWidget
            The shell widget.
        """
        self.get_widget().add_shellwidget(shellwidget)

    def remove_shellwidget(self, shellwidget):
        """
        Remove a registered shellwidget.

        Parameters
        ----------
        shellwidget: spyder.plugins.ipyconsole.widgets.shell.ShellWidget
            The shell widget.
        """
        self.get_widget().remove_shellwidget(shellwidget)

    def add_errored_shellwidget(self, shellwidget):
        """
        Add a new shellwidget whose kernel failed to start.

        Parameters
        ----------
        shellwidget: spyder.plugins.ipyconsole.widgets.shell.ShellWidget
            The shell widget.
        """
        self.get_widget().add_errored_shellwidget(shellwidget)

    def current_widget(self):
        """
        Return the current widget displayed at the moment.

        Returns
        -------
        current_widget: QWidget
            The widget displayed in the current tab.
        """
        return self.get_widget().current_widget()

    def get_widget_for_shellwidget(self, shellwidget):
        """
        Return the widget registered with the given shellwidget.

        Parameters
        ----------
        shellwidget: spyder.plugins.ipyconsole.widgets.shell.ShellWidget
            The shell widget.

        Returns
        -------
        current_widget: QWidget
            The widget corresponding to the shellwidget, or None if not found.
        """
        return self.get_widget().get_widget_for_shellwidget(shellwidget)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Script for running Spyder tests programmatically.
"""

# Standard library imports
import argparse
import os

# To activate/deactivate certain things for pytests only
# NOTE: Please leave this before any other import here!!
os.environ['SPYDER_PYTEST'] = 'True'

# Third party imports
# NOTE: This needs to be imported before any QApplication.
# Don't remove it or change it to a different location!
# pylint: disable=wrong-import-position
from qtpy import QtWebEngineWidgets  # noqa
import pytest


# To run our slow tests only in our CIs
CI = bool(os.environ.get('CI', None))
RUN_SLOW = os.environ.get('RUN_SLOW', None) == 'true'


def run_pytest(run_slow=False, extra_args=None, remoteclient=False):
    """Run pytest tests for Spyder."""
    # Be sure to ignore subrepos and remoteclient plugin
    pytest_args = ['-vv', '-rw', '--durations=10', '--ignore=./external-deps',
                   '-W ignore::UserWarning', '--timeout=120',
                   '--timeout_method=thread']

    if CI:
        # Show coverage
        pytest_args += ['--cov=spyder', '--no-cov-on-fail', '--cov-report=xml']

        # To display nice tests resume in Azure's web page
        if os.environ.get('AZURE', None) is not None:
            pytest_args += ['--cache-clear', '--junitxml=result.xml']
    if run_slow or RUN_SLOW:
        pytest_args += ['--run-slow']
    # Allow user to pass a custom test path to pytest to e.g. run just one test
    if extra_args:
        pytest_args += extra_args

    if remoteclient:
        pytest_args += ['--container-scope=class',
                        './spyder/plugins/remoteclient']
        os.environ["SPYDER_TEST_REMOTE_CLIENT"] = "true"
    else:
        pytest_args += ['--ignore=./spyder/plugins/remoteclient']

    print("Pytest Arguments: " + str(pytest_args))
    errno = pytest.main(pytest_args)

    # sys.exit doesn't work here because some things could be running in the
    # background (e.g. closing the main window) when this point is reached.
    # If that's the case, sys.exit doesn't stop the script as you would expect.
    if errno != 0:
        raise SystemExit(errno)


def main():
    """Parse args then run the pytest suite for Spyder."""
    test_parser = argparse.ArgumentParser(
        usage=('python runtests.py'
               '[-h] [--run-slow] [--remote-client] [pytest_args]'),
        description="Helper script to run Spyder's test suite")
    test_parser.add_argument('--run-slow', action='store_true', default=False,
                             help='Run the slow tests')
    test_parser.add_argument("--remote-client", action="store_true",
                             default=False, help="Run the remote client tests")
    test_args, pytest_args = test_parser.parse_known_args()
    run_pytest(run_slow=test_args.run_slow, extra_args=pytest_args,
               remoteclient=test_args.remote_client)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
"""
Command-line interface for menuinst in Spyder installer
"""

import sys
from pathlib import Path
from argparse import ArgumentParser

from menuinst import api

_base = Path(sys.prefix)
_target = _base / "envs/spyder-runtime"
_menu = _target / "Menu/spyder-menu.json"


def shortcut(**args):
    _, menu_items = api._load(**args)
    print(menu_items[0]._paths()[0])


# ---- Parser for common option items
_parser = ArgumentParser()
_parser.add_argument(
    "--menu",
    dest="metadata_or_path",
    default=str(_menu),
    metavar="PATH",
    type=lambda x: str(Path(x).absolute()),
    help="Path to menu.json file"
)
_parser.add_argument(
    "--target",
    dest="target_prefix",
    default=str(_target),
    type=lambda x: str(Path(x).absolute()),
    help="Path to target environment prefix"
)
_parser.add_argument(
    "--base",
    dest="base_prefix",
    default=str(_base),
    type=lambda x: str(Path(x).absolute()),
    help="Path to base environment prefix"
)

# ---- Main parser for menuinst_cli
parser = ArgumentParser(
    description="Command-line interface for menuinst in Spyder installer"
)
subparsers = parser.add_subparsers(
    title="subcommands",
    required=True,
)

# ---- Subcommand parsers
parser_install = subparsers.add_parser(
    "install", help="Install shortcut.",
    description="Install shortcut.",
    parents=[_parser],
    conflict_handler="resolve",
)
parser_install.set_defaults(func=api.install)

parser_remove = subparsers.add_parser(
    "remove", help="Remove shortcut.",
    description="Remove shortcut.",
    parents=[_parser],
    conflict_handler="resolve",
)
parser_remove.set_defaults(func=api.remove)

parser_shortcut = subparsers.add_parser(
    "shortcut", help="Print shortcut path.",
    description="Print shortcut path.",
    parents=[_parser],
    conflict_handler="resolve",
)
parser_shortcut.add_argument(
    "--mode", dest="_mode", choices=["system", "user"], default="user",
    help="Whether to perform system or user install."
)
parser_shortcut.set_defaults(func=shortcut)

args = vars(parser.parse_args())
func = args.pop("func")
func(**args)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2021, Spyder Bot
#
# Licensed under the terms of the MIT license
# ----------------------------------------------------------------------------
"""
Spyder Boilerplate Plugin.
"""

# Third party imports
import qtawesome as qta
from qtpy.QtWidgets import QHBoxLayout, QTextEdit

# Spyder imports
from spyder.api.config.decorators import on_conf_change
from spyder.api.plugins import SpyderDockablePlugin
from spyder.api.preferences import PluginConfigPage
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.plugins.layout.layouts import VerticalSplitLayout2
from spyder.utils.palette import SpyderPalette


class SpyderBoilerplateConfigPage(PluginConfigPage):

    # --- PluginConfigPage API
    # ------------------------------------------------------------------------
    def setup_page(self):
        pass


class SpyderBoilerplateActions:
    ExampleAction = "example_action"


class SpyderBoilerplateToolBarSections:
    ExampleSection = "example_section"


class SpyderBoilerplateOptionsMenuSections:
    ExampleSection = "example_section"


class SpyderBoilerplateWidget(PluginMainWidget):

    # PluginMainWidget class constants

    # Signals

    def __init__(self, name=None, plugin=None, parent=None):
        super().__init__(name, plugin, parent)

        # Create example widgets
        self._example_widget = QTextEdit(self)
        self._example_widget.setText("Example text")

        # Add example label to layout
        layout = QHBoxLayout()
        layout.addWidget(self._example_widget)
        self.setLayout(layout)

    # --- PluginMainWidget API
    # ------------------------------------------------------------------------
    def get_title(self):
        return "Spyder boilerplate plugin"

    def get_focus_widget(self):
        return self

    def setup(self):
        # Create an example action
        example_action = self.create_action(
            name=SpyderBoilerplateActions.ExampleAction,
            text="Example action",
            tip="Example hover hint",
            icon=self.create_icon("python"),
            triggered=lambda: print("Example action triggered!"),
        )

        # Add an example action to the plugin options menu
        menu = self.get_options_menu()
        self.add_item_to_menu(
            example_action,
            menu,
            SpyderBoilerplateOptionsMenuSections.ExampleSection,
        )

        # Add an example action to the plugin toolbar
        toolbar = self.get_main_toolbar()
        self.add_item_to_toolbar(
            example_action,
            toolbar,
            SpyderBoilerplateOptionsMenuSections.ExampleSection,
        )

        # Shortcuts
        self.register_shortcut_for_widget(
            "Change text",
            self.change_text,
        )

        self.register_shortcut_for_widget(
            "new text",
            self.new_text,
            context="editor",
            plugin_name=self._plugin.NAME,
        )

    def update_actions(self):
        pass

    @on_conf_change
    def on_section_conf_change(self, section):
        pass

    # --- Public API
    # ------------------------------------------------------------------------
    def change_text(self):
        if self._example_widget.toPlainText() == "":
            self._example_widget.setText("Example text")
        else:
            self._example_widget.setText("")

    def new_text(self):
        if self._example_widget.toPlainText() != "Another text":
            self._example_widget.setText("Another text")


class SpyderBoilerplate(SpyderDockablePlugin):
    """
    Spyder Boilerplate plugin.
    """

    NAME = "spyder_boilerplate"
    REQUIRES = []
    OPTIONAL = []
    WIDGET_CLASS = SpyderBoilerplateWidget
    CONF_SECTION = NAME
    CONF_WIDGET_CLASS = SpyderBoilerplateConfigPage
    CUSTOM_LAYOUTS = [VerticalSplitLayout2]
    CONF_DEFAULTS = [
        (CONF_SECTION, {}),
        (
            "shortcuts",
            # Note: These shortcut names are capitalized to check we can
            # set/get/reset them correctly.
            {f"{NAME}/Change text": "Ctrl+B", "editor/New text": "Ctrl+H"},
        ),
    ]

    # --- Signals

    # --- SpyderDockablePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return "Spyder boilerplate plugin"

    @staticmethod
    def get_description():
        return "A boilerplate plugin for testing."

    @staticmethod
    def get_icon():
        return qta.icon('mdi6.alpha-b-box', color=SpyderPalette.ICON_1)

    def on_initialize(self):
        pass

    def check_compatibility(self):
        valid = True
        message = ""  # Note: Remember to use _("") to localize the string
        return valid, message

    def on_close(self, cancellable=True):
        return True

    # --- Public API
    # ------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for EditorStack keyboard shortcuts.
"""

# Standard library imports
import os
import sys
from unittest.mock import Mock

# Third party imports
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.editor.widgets.gotoline import GoToLineDialog
from spyder.plugins.editor.widgets.editorstack import EditorStack


# ---- Qt Test Fixtures
@pytest.fixture
def editorstack(qtbot):
    """
    Set up EditorStack with CodeEditors containing some Python code.
    The cursor is at the empty line below the code.
    """
    editorstack = EditorStack(None, [], False)
    editorstack.set_find_widget(Mock())
    editorstack.set_io_actions(Mock(), Mock(), Mock(), Mock())
    editorstack.close_split_action.setEnabled(False)
    editorstack.new('foo.py', 'utf-8', 'Line1\nLine2\nLine3\nLine4')

    qtbot.addWidget(editorstack)
    editorstack.show()
    editorstack.go_to_line(1)

    # We need to wait for a bit so shortcuts are registered correctly
    qtbot.wait(300)

    return editorstack


# ---- Tests
def test_default_keybinding_values():
    """
    Assert that the default Spyder keybindings for the keyboard shortcuts
    are as expected. This is required because we do not use the keybindings
    saved in Spyder's config to simulate the user keyboard action, due to
    the fact that it is complicated to convert and pass reliably a sequence
    of key strings to qtbot.keyClicks.
    """
    # Assert default keybindings.
    assert CONF.get_shortcut('editor', 'start of document') == 'Ctrl+Home'
    assert CONF.get_shortcut('editor', 'end of document') == 'Ctrl+End'
    assert CONF.get_shortcut('editor', 'delete') == 'Del'
    assert CONF.get_shortcut('editor', 'undo') == 'Ctrl+Z'
    assert CONF.get_shortcut('editor', 'redo') == 'Ctrl+Shift+Z'
    assert CONF.get_shortcut('editor', 'copy') == 'Ctrl+C'
    assert CONF.get_shortcut('editor', 'paste') == 'Ctrl+V'
    assert CONF.get_shortcut('editor', 'cut') == 'Ctrl+X'
    assert CONF.get_shortcut('editor', 'select all') == 'Ctrl+A'
    assert CONF.get_shortcut('editor', 'delete line') == 'Ctrl+D'
    assert CONF.get_shortcut('editor', 'transform to lowercase') == 'Ctrl+U'
    assert CONF.get_shortcut('editor',
                             'transform to uppercase') == 'Ctrl+Shift+U'
    assert CONF.get_shortcut('editor', 'go to line') == 'Ctrl+L'
    assert CONF.get_shortcut('editor', 'next word') == 'Ctrl+Right'
    assert CONF.get_shortcut('editor', 'previous word') == 'Ctrl+Left'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
def test_start_and_end_of_document_shortcuts(editorstack, qtbot):
    """
    Test that the start of document and end of document shortcut are working
    as expected.
    """
    editor = editorstack.get_current_editor()

    # Assert initial state.
    assert editor.get_cursor_line_column() == (0, 0)

    # Go to the end of the document.
    qtbot.keyClick(editor, Qt.Key_End, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (4, 0)

    # Go to the start of the document.
    qtbot.keyClick(editor, Qt.Key_Home, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (0, 0)


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
def test_del_undo_redo_shortcuts(editorstack, qtbot):
    """
    Test that the undo and redo keyboard shortcuts are working as expected
    with the default Spyder keybindings.

    Regression test for spyder-ide/spyder#7743.
    """
    editor = editorstack.get_current_editor()

    # Delete the first character of the first line.
    qtbot.keyClick(editor, Qt.Key_Delete)
    assert editor.toPlainText() == 'ine1\nLine2\nLine3\nLine4\n'

    # Undo the last action with Ctrl+Z.
    qtbot.keyClick(editor, Qt.Key_Z, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == 'Line1\nLine2\nLine3\nLine4\n'

    # Redo the last action with Ctrl+Shift+Z .
    qtbot.keyClick(editor, Qt.Key_Z,
                   modifier=Qt.ControlModifier | Qt.ShiftModifier)
    assert editor.toPlainText() == 'ine1\nLine2\nLine3\nLine4\n'

    # Undo the last action again with Ctrl+Z.
    qtbot.keyClick(editor, Qt.Key_Z, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == 'Line1\nLine2\nLine3\nLine4\n'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
def test_copy_cut_paste_shortcuts(editorstack, qtbot):
    """
    Test that the copy, cut, and paste keyboard shortcuts are working as
    expected with the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()
    QApplication.clipboard().clear()

    # Select and Copy the first line in the editor.
    qtbot.keyClick(editor, Qt.Key_End, modifier=Qt.ShiftModifier)
    assert editor.get_selected_text() == 'Line1'

    qtbot.keyClick(editor, Qt.Key_C, modifier=Qt.ControlModifier)
    assert QApplication.clipboard().text() == 'Line1'

    # Paste the selected text.
    qtbot.keyClick(editor, Qt.Key_Home)
    qtbot.keyClick(editor, Qt.Key_V, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == 'Line1Line1\nLine2\nLine3\nLine4\n'

    # Select and Cut the first line in the editor.
    qtbot.keyClick(editor, Qt.Key_Home)
    qtbot.keyClick(editor, Qt.Key_End, modifier=Qt.ShiftModifier)
    assert editor.get_selected_text() == 'Line1Line1'

    qtbot.keyClick(editor, Qt.Key_X, modifier=Qt.ControlModifier)
    assert QApplication.clipboard().text() == 'Line1Line1'
    assert editor.toPlainText() == '\nLine2\nLine3\nLine4\n'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
def test_select_all_shortcut(editorstack, qtbot):
    """
    Test that the select all keyboard shortcut is working as
    expected with the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()

    # Select all the text in the editor.
    qtbot.keyClick(editor, Qt.Key_A, modifier=Qt.ControlModifier)
    assert editor.get_selected_text() == 'Line1\nLine2\nLine3\nLine4\n'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
@pytest.mark.no_xvfb
def test_delete_line_shortcut(editorstack, qtbot):
    """
    Test that the delete line keyboard shortcut is working as
    expected with the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()

    # Delete the second line in the editor.
    editor.go_to_line(2)
    qtbot.keyClick(editor, Qt.Key_D, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == 'Line1\nLine3\nLine4\n'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
@pytest.mark.no_xvfb
def test_go_to_line_shortcut(editorstack, qtbot, mocker):
    """
    Test that the go to line keyboard shortcut is working
    as expected with the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()
    qtbot.keyClick(editor, Qt.Key_Home, modifier=Qt.ControlModifier)

    # Go to line 3.
    mocker.patch.object(GoToLineDialog, 'exec_', return_value=True)
    mocker.patch.object(GoToLineDialog, 'get_line_number', return_value=3)
    qtbot.keyClick(editor, Qt.Key_L, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (2, 0)


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
@pytest.mark.no_xvfb
def test_transform_to_lowercase_shortcut(editorstack, qtbot):
    """
    Test that the transform to lowercase shorcut is working as expected with
    the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()

    # Transform all the text to lowercase.
    qtbot.keyClick(editor, Qt.Key_A, modifier=Qt.ControlModifier)
    qtbot.keyClick(editor, Qt.Key_U, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == 'line1\nline2\nline3\nline4\n'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
@pytest.mark.no_xvfb
def test_transform_to_uppercase_shortcut(editorstack, qtbot):
    """
    Test that the transform to uppercase shortcuts is working as expected with
    the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()

    # Transform all the text to uppercase.
    qtbot.keyClick(editor, Qt.Key_A, modifier=Qt.ControlModifier)
    qtbot.keyClick(editor, Qt.Key_U,
                   modifier=Qt.ControlModifier | Qt.ShiftModifier)
    assert editor.toPlainText() == 'LINE1\nLINE2\nLINE3\nLINE4\n'


@pytest.mark.skipif(
    sys.platform.startswith('linux') and running_in_ci(),
    reason="It fails on Linux due to the lack of a proper X server.")
@pytest.mark.no_xvfb
def test_next_and_previous_word_shortcuts(editorstack, qtbot):
    """
    Test that the next word and previous word shortcuts are working as
    expected with the default Spyder keybindings.
    """
    editor = editorstack.get_current_editor()

    # Go to the next word 3 times.
    assert editor.get_cursor_line_column() == (0, 0)
    qtbot.keyClick(editor, Qt.Key_Right, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (0, 5)
    qtbot.keyClick(editor, Qt.Key_Right, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (1, 0)
    qtbot.keyClick(editor, Qt.Key_Right, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (1, 5)

    # Go to the previous word 3 times.
    qtbot.keyClick(editor, Qt.Key_Left, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (1, 0)
    qtbot.keyClick(editor, Qt.Key_Left, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (0, 5)
    qtbot.keyClick(editor, Qt.Key_Left, modifier=Qt.ControlModifier)
    assert editor.get_cursor_line_column() == (0, 0)


@pytest.mark.skipif(sys.platform == 'darwin', reason="Not valid in macOS")
def test_builtin_shift_del_and_ins(editorstack, qtbot):
    """
    Test that the builtin key sequences Ctrl+Ins, Shit+Del and Shift+Ins result
    in copy, cut and paste actions in Windows and Linux.

    Regression test for spyder-ide/spyder#5035, spyder-ide/spyder#4947, and
    spyder-ide/spyder#5973.
    """
    editor = editorstack.get_current_editor()
    QApplication.clipboard().clear()

    # Select the first line of the editor.
    qtbot.keyClick(editor, Qt.Key_End, modifier=Qt.ShiftModifier)
    assert editor.get_selected_text() == 'Line1'

    # Copy the selection with Ctrl+Ins.
    qtbot.keyClick(editor, Qt.Key_Insert, modifier=Qt.ControlModifier)
    assert QApplication.clipboard().text() == 'Line1'

    # Paste the copied text at the end of the line with Shift+Ins.
    qtbot.keyClick(editor, Qt.Key_End)
    qtbot.keyClick(editor, Qt.Key_Insert, modifier=Qt.ShiftModifier)
    assert editor.toPlainText() == 'Line1Line1\nLine2\nLine3\nLine4\n'

    # Select the second line in the editor again.
    qtbot.keyClick(editor, Qt.Key_Home, modifier=Qt.ShiftModifier)
    assert editor.get_selected_text() == 'Line1Line1'

    # Cut the selection with Shift+Del.
    qtbot.keyClick(editor, Qt.Key_Delete, modifier=Qt.ShiftModifier)
    assert QApplication.clipboard().text() == 'Line1Line1'
    assert editor.toPlainText() == '\nLine2\nLine3\nLine4\n'


@pytest.mark.skipif(os.name != 'nt', reason='Only valid in Windows system')
def test_builtin_undo_redo(editorstack, qtbot):
    """
    Test that the builtin key sequence Alt+Backspace, Ctrl+Y and
    Alt+Shift+Backspace result in, respectively, an undo, redo and redo action
    in Windows.
    """
    editor = editorstack.get_current_editor()

    # Write something on a new line.
    qtbot.keyClicks(editor, 'Something')
    qtbot.keyClick(editor, Qt.Key_Return)
    assert editor.toPlainText() == 'Something\nLine1\nLine2\nLine3\nLine4\n'

    # Undo the last action with Alt+Backspace.
    qtbot.keyClick(editor, Qt.Key_Backspace, modifier=Qt.AltModifier)
    assert editor.toPlainText() == 'SomethingLine1\nLine2\nLine3\nLine4\n'

    # Undo the second to last action with Alt+Backspace.
    qtbot.keyClick(editor, Qt.Key_Backspace, modifier=Qt.AltModifier)
    assert editor.toPlainText() == 'Line1\nLine2\nLine3\nLine4\n'

    # Redo the second to last action with Alt+Shift+Backspace.
    qtbot.keyClick(editor, Qt.Key_Backspace,
                   modifier=Qt.AltModifier | Qt.ShiftModifier)
    assert editor.toPlainText() == 'SomethingLine1\nLine2\nLine3\nLine4\n'

    # Redo the last action with Ctrl+Y.
    qtbot.keyClick(editor, Qt.Key_Y, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == 'Something\nLine1\nLine2\nLine3\nLine4\n'


@pytest.mark.skipif(
    sys.platform.startswith("linux") and running_in_ci(),
    reason='Fails on Linux and CI'
)
def test_shortcuts_for_new_editors(editorstack, qtbot):
    """
    Test that widget shortcuts are working for new editors.

    This is a regression test for spyder-ide/spyder#23151.
    """
    # Create new file and give it focus
    editorstack.new('bar.py', 'utf-8', 'Line5\nLine6\nLine7\nLine8')
    editorstack.tabs.setCurrentIndex(1)

    # Go to its first line, click the shortcut to comment it and check it was
    editorstack.go_to_line(1)
    editor = editorstack.get_current_editor()
    qtbot.keyClick(editor, Qt.Key_1, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == '# Line5\nLine6\nLine7\nLine8\n'


if __name__ == "__main__":
    import os
    pytest.main(['-x', os.path.basename(__file__), '-vv', '-rw'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Run API tests."""

# Standard library imports
import copy
from datetime import datetime
from itertools import chain, repeat
import logging
from types import MethodType
from typing import Dict, Tuple, Optional, List
from uuid import uuid4
from unittest.mock import MagicMock

# Third-part imports
import pytest
from qtpy.QtCore import Signal, Qt
from qtpy.QtWidgets import (
    QAction, QWidget, QCheckBox, QLineEdit, QVBoxLayout, QHBoxLayout, QLabel)

# Local imports
from spyder.plugins.run.api import (
    RunExecutor, RunConfigurationProvider, RunConfigurationMetadata, Context,
    RunConfiguration, SupportedExtensionContexts, RunExecutionParameters,
    RunExecutorConfigurationGroup, ExtendedRunExecutionParameters,
    PossibleRunResult, RunContext, ExtendedContext, RunActions, run_execute,
    WorkingDirOpts, WorkingDirSource, StoredRunExecutorParameters)
from spyder.plugins.run.plugin import Run


logger = logging.getLogger(__name__)


class MockedMainWindow(QWidget, MagicMock):
    def get_plugin(self, name, error=True):
        return MagicMock()


class ExampleConfigurationProvider(RunConfigurationProvider):
    sig_uuid_focus_requested = Signal(str)

    def __init__(self, parent, info, provider_name) -> None:
        super().__init__(parent)

        # self.executor_configuration = []
        self.current_uuid = None
        self.provider_name = provider_name
        self.run_configurations: Dict[str, RunConfigurationMetadata] = {}
        self.configuration_uuids: Dict[Tuple[str, str], str] = {}
        self.supported_extensions: List[SupportedExtensionContexts] = []
        self.actions: Dict[str, QAction] = {}

        for ext, context, reg, create_btn in info:
            ctx = Context(name=context)
            ext_ctx = ExtendedContext(context=ctx, is_super=reg)

            self.supported_extensions.append(
                SupportedExtensionContexts(
                    input_extension=ext, contexts=[ext_ctx]))

            if reg:
                uuid = str(uuid4())
                metadata = RunConfigurationMetadata(
                    name=f'{ext}_{context.lower()}_{provider_name}_example',
                    source=provider_name,
                    path=f'test-example-{context}.{ext}',
                    timestamp=datetime.now(), uuid=uuid,
                    context=ctx,
                    input_extension=ext)

                self.run_configurations[uuid] = metadata
                self.configuration_uuids[(ext, context)] = uuid

            if create_btn:
                self.actions[context] = None

    def on_run_available(self, run: Run):
        self.sig_uuid_focus_requested.connect(
            run.switch_focused_run_configuration
        )

        run.register_run_configuration_provider(
            self.provider_name, self.supported_extensions)

        for uuid in self.run_configurations:
            metadata = self.run_configurations[uuid]
            run.register_run_configuration_metadata(self, metadata)

        for context_name in list(self.actions):
            act = run.create_run_button(
                getattr(RunContext, context_name),
                f'Run {context_name}',
                icon=None,
                tip=None,
                shortcut_context=None,
                register_shortcut=False,
                add_to_toolbar=False,
                add_to_menu=False
            )

            self.actions[context_name] = act

    def on_run_teardown(self, run):
        run.deregister_run_configuration_provider(
            self.provider_name, self.supported_extensions)

        for uuid in self.run_configurations:
            # metadata = self.run_configurations[uuid]
            run.deregister_run_configuration_metadata(uuid)

        for context_name in list(self.actions):
            run.destroy_run_button(getattr(RunContext, context_name))
            self.actions.pop(context_name)

    def get_run_configuration(self, uuid: str) -> RunConfiguration:
        metadata = self.run_configurations[uuid]
        name = metadata['name']
        source = metadata['source']
        path = metadata['path']
        context = metadata['context']['name']
        extension = metadata['input_extension']
        run_input = {
            'contents': (f'File: {name} | Location: {path} | Source: {source} '
                         f'| Context: {context} | Ext: {extension}')
        }

        return RunConfiguration(run_input=run_input, metadata=metadata)

    def get_run_configuration_per_context(
        self,
        context: str,
        action_name: Optional[str] = None,
        context_modificator: Optional[str] = None,
        re_run: bool = False
    ) -> Optional[RunConfiguration]:

        metadata = copy.deepcopy(self.run_configurations[self.current_uuid])
        inverse_run_ctx = {RunContext[k]: k for k in RunContext}
        metadata['context'] = Context(name=inverse_run_ctx[context],
                                      identifier=context)

        name = metadata['name']
        source = metadata['source']
        path = metadata['path']
        context = metadata['context']['name']
        extension = metadata['input_extension']
        run_input = {
            'contents': (f'File: {name} | Location: {path} | Source: {source} '
                         f'| Context: {context} | Ext: {extension} '
                         f'| Action: {action_name} | Re-run: {re_run}'),
            'action': action_name,
            're_run': re_run
        }

        return RunConfiguration(run_input=run_input, metadata=metadata)

    def switch_focus(self, ext: str, context: str):
        uuid = None
        if (ext, context) in self.configuration_uuids:
            uuid = self.configuration_uuids[(ext, context)]
        self.current_uuid = uuid
        self.sig_uuid_focus_requested.emit(uuid)

    def focus_run_configuration(self, uuid: str):
        metadata = self.run_configurations[uuid]
        context = metadata['context']['name']
        extension = metadata['input_extension']
        self.switch_focus(extension, context)


class GenExampleRunExecutorConf(RunExecutorConfigurationGroup):

    def __init__(self, parent: QWidget, context: Context, input_extension: str,
                 input_metadata: RunConfigurationMetadata):
        super().__init__(parent, context, input_extension, input_metadata)
        default_conf = self.get_default_configuration()
        self.widgets = {}

        layout = QVBoxLayout()
        for key_name in default_conf:
            default_value = default_conf[key_name]
            if isinstance(default_value, bool):
                widget = QCheckBox(key_name)
                layout.addWidget(widget)
            elif isinstance(default_value, str):
                temp_layout = QHBoxLayout()
                label = QLabel(key_name)
                temp_layout.addWidget(label)
                widget = QLineEdit()
                temp_layout.addWidget(widget)
                layout.addLayout(temp_layout)
            self.widgets[key_name] = widget

        self.setLayout(layout)

    def get_configuration(self) -> dict:
        conf = {}
        for key_name in self.widgets:
            widget = self.widgets[key_name]
            if isinstance(widget, QCheckBox):
                conf[key_name] = widget.isChecked()
            elif isinstance(widget, QLineEdit):
                conf[key_name] = widget.text()
        return conf

    def set_configuration(self, config: dict):
        for key_name in config:
            value = config[key_name]
            widget = self.widgets[key_name]
            if isinstance(widget, QCheckBox):
                widget.setChecked(value)
            elif isinstance(widget, QLineEdit):
                widget.setText(value)


class MetaExampleRunExecutorConf(type(GenExampleRunExecutorConf)):

    def __new__(cls, clsname, bases, attrs):
        default_args = attrs.pop('default_args_meta')

        def get_default_configuration() -> dict:
            return default_args

        return super(MetaExampleRunExecutorConf, cls).__new__(
            cls, clsname, bases, {
                **attrs,
                'get_default_configuration': staticmethod(
                    get_default_configuration)
            }
        )


def ExampleRunExecutorConfFactory(default_args_dict: dict):

    class WrappedExampleRunExecutorConf(
            GenExampleRunExecutorConf, metaclass=MetaExampleRunExecutorConf):
        default_args_meta = default_args_dict

    return WrappedExampleRunExecutorConf


def gen_executor_handler(executor_name, handler_name, ext=None, context=None):

    @run_execute(extension=ext, context=context)
    def executor_handler(
            self, input: RunConfiguration,
            conf: ExtendedRunExecutionParameters) -> List[PossibleRunResult]:
        self.sig_run_invocation.emit(
            (executor_name, handler_name, input, conf))

    return executor_handler


class ExampleRunExecutorWrapper(RunExecutor):

    sig_run_invocation = Signal(tuple)

    def __init__(self, parent, info, executor_name):
        self.executor_configuration = []
        self.handlers = {}
        self.actions = {}
        self.executor_name = executor_name
        self.handlers['all'] = self.bind_execution_method('all')

        for ext, context, prio, default_conf, req_cwd, handler, btn in info:
            ConfWidget = ExampleRunExecutorConfFactory(default_conf)
            context_id = RunContext[context]
            self.executor_configuration.append({
                'input_extension': ext,
                'context': {
                    'name': context
                },
                'output_formats': [],
                'configuration_widget': ConfWidget,
                'requires_cwd': req_cwd,
                'priority': prio
            })

            if handler == 'context':
                self.handlers[context] = self.bind_execution_method(
                    f'context_{context_id}', context=context_id)
            elif handler == 'ext':
                self.handlers[ext] = self.bind_execution_method(
                    f'ext_{ext}', ext=ext)
            elif handler == 'both':
                self.handlers[(ext, context_id)] = self.bind_execution_method(
                    f'{ext}_{context_id}', ext=ext, context=context_id)

            if btn:
                self.actions[context_id] = None

        super().__init__(parent)

    def bind_execution_method(self, handler_name, ext=None, context=None):
        func = gen_executor_handler(
            self.executor_name, handler_name, ext, context)
        meth = MethodType(func, self)
        setattr(self, f'exec_{handler_name}', meth)
        return meth

    def on_run_available(self, run):
        run.register_executor_configuration(self, self.executor_configuration)
        for context_id in list(self.actions):
            act = run.create_run_in_executor_button(
                context_id,
                self.NAME,
                text=f'Run {context_id} in {self.NAME}',
                tip=None,
                icon=None,
                shortcut_context=None,
                register_shortcut=False,
                add_to_menu=False
            )
            self.actions[context_id] = act

    def on_run_teardown(self, run):
        run.deregister_executor_configuration(self, self.executor_configuration)
        for context_id in list(self.actions):
            run.destroy_run_in_executor_button(context_id, self.NAME)
            self.actions.pop(context_id)


class MetaExampleRunExecutor(type(ExampleRunExecutorWrapper)):

    def __new__(cls, clsname, bases, attrs):
        executor_name = attrs.pop('executor_name_meta')

        def get_name():
            return executor_name

        return super(MetaExampleRunExecutor, cls).__new__(
            cls, clsname, bases, {
                **attrs,
                'get_name': staticmethod(get_name),
                'NAME': executor_name
            }
        )


def ExampleRunExecutorFactory(parent, info, executor_name):

    class ExampleRunExecutor(
            ExampleRunExecutorWrapper, metaclass=MetaExampleRunExecutor):
        executor_name_meta = executor_name

    return ExampleRunExecutor(parent, info, executor_name)


@pytest.fixture
def run_mock(qtbot, tmpdir):
    temp_dir = str(tmpdir.mkdir('run'))
    mock_main_window = MockedMainWindow()
    run = Run(mock_main_window, None)
    run.on_initialize()
    run._switch_working_dir(temp_dir)
    return run, mock_main_window, temp_dir


def test_run_plugin(qtbot, run_mock):
    run, main_window, temp_cwd = run_mock

    # Create mock run configuration providers
    provider_1_conf = [
        ('ext1', 'RegisteredContext', True, False),
        ('ext1', 'SubordinateContext1', False, True),
        ('ext1', 'UnusedContext', False, True),
        ('ext2', 'AnotherSuperContext', True, False),
        ('ext3', 'RegisteredContext', True, False)
    ]

    provider_2_conf = [
        ('ext1', 'RegisteredContext', True, False),
        ('ext1', 'SubordinateContext1', False, True),
        ('ext1', 'SubordinateContext2', False, True),
        ('ext3', 'AnotherSuperContext', True, False)
    ]

    exec_provider_1 = ExampleConfigurationProvider(
        main_window, provider_1_conf, 'conf_prov_1')

    exec_provider_2 = ExampleConfigurationProvider(
        main_window, provider_2_conf, 'conf_prov_2')

    # Register providers with the Run plugin
    exec_provider_1.on_run_available(run)
    exec_provider_2.on_run_available(run)

    # Assert that the actions for 'SubordinateContext1' are the same for both
    # providers.
    act1 = exec_provider_1.actions['SubordinateContext1']
    act2 = exec_provider_2.actions['SubordinateContext1']
    assert act1 == act2

    # Create mock run executors
    executor_1_conf = [
        (
            'ext1', 'RegisteredContext', 0, {
                'opt1': True,
                'opt2': '',
                'opt3': False
            }, True, 'both', True
        ),
        (
            'ext1', 'SubordinateContext1', 1, {
                'arg1': '',
                'arg2': False,
                'arg3': False
            }, False, 'ext', True
        ),
        (
            'ext2', 'AnotherSuperContext', 0, {
                'only_opt': '',
            }, True, 'context', False
        ),
        (
            'ext3', 'AnotherSuperContext', 0, {
                'name_1': False,
                'name_2': True
            }, False, 'context', False
        ),
    ]

    executor_2_conf = [
        (
            'ext1', 'RegisteredContext', 1, {
                'ex2_opt1': True,
                'ex2_opt2': False,
                'ex2_opt3': False,

            }, True, 'ext', True
        ),
        (
            'ext1', 'SubordinateContext1', 0, {
                'arg1': '',
                'arg2': False,
                'arg3': False
            }, True, 'ext', True
        ),
        (
            'ext1', 'SubordinateContext2', 0, {
                'opt1': True,
                'opt2': False,
                'opt3': False
            }, True, 'context', True
        ),
        (
            'ext3', 'RegisteredContext', 0, {
                'name_1': False,
                'name_2': True
            }, False, 'all', False
        ),
    ]

    executor_1 = ExampleRunExecutorFactory(
        main_window, executor_1_conf, 'executor_1')
    executor_2 = ExampleRunExecutorFactory(
        main_window, executor_2_conf, 'executor_2')

    # Register run executors on the Run plugin
    executor_1.on_run_available(run)
    executor_2.on_run_available(run)

    # Focus on the first run configuration for the first configuration provider
    exec_provider_1.switch_focus('ext1', 'RegisteredContext')

    # Assert that both provider and run are in sync
    container = run.get_container()
    run_uuid = container.currently_selected_configuration
    assert run_uuid == exec_provider_1.current_uuid

    # Assert that run-specific context actions are available or disabled
    act = exec_provider_1.actions['SubordinateContext1']
    assert act.isEnabled()

    act = exec_provider_1.actions['UnusedContext']
    assert not act.isEnabled()

    act = exec_provider_2.actions['SubordinateContext2']
    assert not act.isEnabled()

    # Spawn the configuration dialog
    run_act = run.get_action(RunActions.Run)
    run_act.trigger()

    expected_configurations = []
    total_prov_1_conf = zip(
        repeat(exec_provider_1.provider_name, len(provider_1_conf)),
        provider_1_conf)
    total_prov_2_conf = zip(
        repeat(exec_provider_2.provider_name, len(provider_2_conf)),
        provider_2_conf)
    for provider_name, (ext, context, reg, _) in chain(
            total_prov_1_conf, total_prov_2_conf):
        if reg:
            expected_configurations.append(
                (ext, context,
                 f'{ext}_{context.lower()}_{provider_name}_example'))

    executors_per_conf = {}
    executor_1_conf_id = zip(repeat(executor_1.NAME, len(executor_1_conf)),
                             executor_1_conf)
    executor_2_conf_id = zip(repeat(executor_2.NAME, len(executor_2_conf)),
                             executor_2_conf)
    executor_conf_iter = chain(executor_1_conf_id, executor_2_conf_id)
    for (exec_name, (ext, context, prio,
                     default_conf, req_cwd, handler, _)) in executor_conf_iter:
        conf_executors = executors_per_conf.get((ext, context), [])
        conf_executors.insert(
            prio, (exec_name, default_conf, req_cwd, handler))
        executors_per_conf[(ext, context)] = conf_executors

    dialog = container.dialog
    with qtbot.waitSignal(dialog.finished, timeout=200000):
        conf_combo = dialog.configuration_combo
        exec_combo = dialog.executor_combo
        wdir_group = dialog.wdir_group

        # Ensure that there are 5 registered run configurations
        assert conf_combo.count() == 5

        # Ensure that the combobox contain all available configurations
        for i, (_, _, label) in enumerate(expected_configurations):
            combo_label = conf_combo.itemText(i)
            assert label == combo_label

        # Ensure that the currently selected configuration corresponds to the
        # currently selected one.
        assert conf_combo.currentText() == expected_configurations[0][-1]

        # Ensure that the executor settings are being loaded correctly per
        # run configuration
        for i, (ext, context, _) in enumerate(expected_configurations):
            conf_combo.setCurrentIndex(i)
            available_executors = executors_per_conf[(ext, context)]
            # Assert that the order and the executor configurations are loaded
            # according to the priority order.
            for (j, (executor_name,
                     default_conf,
                     req_cwd, _)) in enumerate(available_executors):
                exec_combo.setCurrentIndex(j)
                current_exec_name = exec_combo.currentText()
                conf_widget = dialog.current_widget
                current_conf = conf_widget.get_configuration()

                # Ensure that the selected executor corresponds to the one
                # defined by the priority order
                assert current_exec_name == executor_name

                # Ensure that the working directory options are enabled or
                # disabled according to the executor settings.
                assert not (wdir_group.isEnabled() ^ req_cwd)

                # Ensure that the executor configuration group widget contains
                # the default options declared.
                assert current_conf == default_conf

        # Select the first configuration again
        conf_combo.setCurrentIndex(0)

        # Change some default options
        conf_widget = dialog.current_widget
        cwd_radio = dialog.cwd_radio
        conf_widget.widgets['opt2'].setText('Test')
        cwd_radio.setChecked(True)

        # Execute the configuration
        buttons = dialog.bbox.buttons()
        run_btn = buttons[2]
        with qtbot.waitSignal(executor_1.sig_run_invocation) as sig:
            qtbot.mouseClick(run_btn, Qt.LeftButton)

    # Verify the selected executor output
    test_executor_name, handler_name, run_input, exec_conf = sig.args[0]
    ext, context, name = expected_configurations[0]
    available_executors = executors_per_conf[(ext, context)]
    executor_name, default_conf, _, handler = available_executors[0]
    conf_copy = copy.deepcopy(default_conf)

    conf_copy['opt2'] = 'Test'
    actual_run_input = run_input['run_input']
    metadata = run_input['metadata']

    name = metadata['name']
    source = metadata['source']
    path = metadata['path']
    context = metadata['context']
    run_conf_uuid = metadata['uuid']
    ext = metadata['input_extension']

    contents = actual_run_input['contents']
    assert contents == (f'File: {name} | Location: {path} | Source: {source} '
                        f'| Context: {context["name"]} | Ext: {ext}')

    assert test_executor_name == executor_name
    assert handler == 'both'

    # Ensure that the executor run configuration was saved
    assert exec_conf['uuid'] is not None
    assert exec_conf['name'] == "Default (custom)"

    # Check that the configuration parameters are the ones defined by the
    # dialog
    params = exec_conf['params']
    working_dir = params['working_dir']
    assert params['executor_params'] == conf_copy
    assert working_dir['source'] == WorkingDirSource.CurrentDirectory
    assert working_dir['path'] == temp_cwd

    # Assert that the run_exec dispatcher works for the specific combination
    assert handler_name == f'{ext}_{context["identifier"]}'

    # Assert that the run configuration gets registered
    stored_run_params = run.get_last_used_executor_parameters(run_conf_uuid)
    current_exec_uuid = exec_conf['uuid']
    assert stored_run_params['executor'] == test_executor_name
    assert stored_run_params['selected'] == current_exec_uuid

    # Spawn the configuration dialog
    run_act = run.get_action(RunActions.Run)
    run_act.trigger()

    dialog = container.dialog
    with qtbot.waitSignal(dialog.finished, timeout=200000):
        # Select the first configuration again
        conf_combo.setCurrentIndex(0)

        # Change some options
        conf_widget = dialog.current_widget
        conf_widget.widgets['opt1'].setChecked(False)

        # Execute the configuration
        buttons = dialog.bbox.buttons()
        run_btn = buttons[2]
        with qtbot.waitSignal(executor_1.sig_run_invocation) as sig:
            qtbot.mouseClick(run_btn, Qt.LeftButton)

    # Assert that changes took effect and that the run executor parameters were
    # saved in the Custom config
    _, _, _, exec_conf = sig.args[0]
    params = exec_conf['params']
    assert params['executor_params']['opt1'] is False
    assert exec_conf['uuid'] == current_exec_uuid

    # Focus into another configuration
    exec_provider_2.switch_focus('ext3', 'AnotherSuperContext')

    # Re-run last configuration
    re_run_act = run.get_action(RunActions.ReRun)

    with qtbot.waitSignal(executor_1.sig_run_invocation) as sig:
        re_run_act.trigger()

    # Assert that the previously ran configuration is the same as the current
    # one
    _, _, run_input, _ = sig.args[0]
    re_run_metadata = run_input['metadata']
    assert re_run_metadata == metadata

    # Run a subordinate context on the same executor
    subordinate_act = exec_provider_1.actions['SubordinateContext1']
    with qtbot.waitSignal(executor_1.sig_run_invocation) as sig:
        subordinate_act.trigger()

    subord_executor_name, handler_name, run_input, exec_conf = sig.args[0]
    available_executors = executors_per_conf[('ext1', 'SubordinateContext1')]
    executor_name, default_conf, _, handler = available_executors[1]

    assert subord_executor_name == executor_name
    assert handler_name == 'ext_ext1'

    actual_run_input = run_input['run_input']
    metadata = run_input['metadata']

    name = metadata['name']
    source = metadata['source']
    path = metadata['path']
    context = metadata['context']
    run_conf_uuid = metadata['uuid']
    ext = metadata['input_extension']

    contents = actual_run_input['contents']
    action = actual_run_input['action']
    re_run = actual_run_input['re_run']
    assert contents == (f'File: {name} | Location: {path} | Source: {source} '
                        f'| Context: {context["name"]} | Ext: {ext} '
                        f'| Action: {action} | Re-run: {re_run}')

    params = exec_conf['params']
    working_dir = params['working_dir']
    assert params['executor_params'] == default_conf
    assert working_dir['source'] == WorkingDirSource.ConfigurationDirectory
    assert working_dir['path'] == ''

    # Run a subordinate context on a different executor
    subordinate_act = exec_provider_2.actions['SubordinateContext2']
    with qtbot.waitSignal(executor_2.sig_run_invocation) as sig:
        subordinate_act.trigger()

    subord_executor_name, handler_name, run_input, exec_conf = sig.args[0]
    available_executors = executors_per_conf[('ext1', 'SubordinateContext2')]
    executor_name, default_conf, _, handler = available_executors[0]

    assert subord_executor_name == executor_name
    assert handler_name == 'context_subordinate_context2'

    actual_run_input = run_input['run_input']
    metadata = run_input['metadata']

    name = metadata['name']
    source = metadata['source']
    path = metadata['path']
    context = metadata['context']
    run_conf_uuid = metadata['uuid']
    ext = metadata['input_extension']

    contents = actual_run_input['contents']
    action = actual_run_input['action']
    re_run = actual_run_input['re_run']
    assert contents == (f'File: {name} | Location: {path} | Source: {source} '
                        f'| Context: {context["name"]} | Ext: {ext} '
                        f'| Action: {action} | Re-run: {re_run}')

    params = exec_conf['params']
    working_dir = params['working_dir']
    assert params['executor_params'] == default_conf
    assert working_dir['source'] == WorkingDirSource.ConfigurationDirectory
    assert working_dir['path'] == ''

    # Invoke another executor using one of the dedicated actions
    with qtbot.waitSignal(executor_2.sig_run_invocation) as sig:
        subordinate_act.trigger()

    used_executor_name, handler_name, run_input, exec_conf = sig.args[0]
    available_executors = executors_per_conf[('ext1', 'SubordinateContext1')]
    executor_name, default_conf, _, handler = available_executors[0]

    assert used_executor_name == executor_name

    # Switch to other run configuration and register a set of custom run
    # executor parameters
    exec_provider_2.switch_focus('ext3', 'AnotherSuperContext')
    dialog.exec_()

    # Spawn the configuration dialog
    run_act = run.get_action(RunActions.Run)
    run_act.trigger()

    dialog = container.dialog
    with qtbot.waitSignal(dialog.finished, timeout=200000):
        conf_combo = dialog.configuration_combo
        exec_combo = dialog.executor_combo
        name_params_text = dialog.name_params_text

        # Modify some options
        conf_widget = dialog.current_widget
        conf_widget.widgets['name_1'].setChecked(True)
        conf_widget.widgets['name_2'].setChecked(False)

        # Make sure that the custom configuration is stored
        name_params_text.setText('CustomParams')

        # Execute the configuration
        buttons = dialog.bbox.buttons()
        run_btn = buttons[2]
        with qtbot.waitSignal(executor_1.sig_run_invocation) as sig:
            qtbot.mouseClick(run_btn, Qt.LeftButton)

    executor_name, _, run_input, exec_conf = sig.args[0]

    # Ensure that the configuration got saved
    saved_params = run.get_executor_configuration_parameters(
        executor_name, 'ext3', RunContext.AnotherSuperContext)
    executor_params = saved_params['params']
    assert len(executor_params) == 2

    # Check that the configuration passed to the executor is the same that was
    # saved.
    exec_conf_uuid = exec_conf['uuid']
    current_saved_params = executor_params[exec_conf_uuid]
    current_saved_params['params']['working_dir']['path'] = ''
    assert current_saved_params == exec_conf
    assert current_saved_params['name'] == 'CustomParams'

    # Check that the run configuration points to the latest saved parameters
    metadata = run_input['metadata']
    run_conf_uuid = metadata['uuid']
    stored_run_params = run.get_last_used_executor_parameters(run_conf_uuid)
    assert stored_run_params['executor'] == executor_name
    assert stored_run_params['selected'] == exec_conf_uuid

    # Check deleting a parameters config
    current_exec_params = container.get_conf('parameters')[executor_name]
    current_ext_ctx_params = (
        current_exec_params[('ext3', RunContext.AnotherSuperContext)]['params']
    )
    assert current_ext_ctx_params != {}  # Check params to delete are present

    container.delete_executor_configuration_parameters(
        executor_name, 'ext3', RunContext.AnotherSuperContext, exec_conf_uuid
    )

    new_exec_params = container.get_conf('parameters')[executor_name]
    new_ext_ctx_params = (
        new_exec_params[('ext3', RunContext.AnotherSuperContext)]['params']
    )
    assert len(new_ext_ctx_params) == 1

    # Check that adding new parameters preserves the previous ones
    current_exec_params = container.get_conf('parameters')[executor_name]
    assert (
        len(
            current_exec_params[("ext1", RunContext.RegisteredContext)][
                "params"
            ]
        )
        == 2
    )  # Check that we have one config in this context

    new_exec_conf_uuid = str(uuid4())
    new_params = StoredRunExecutorParameters(
        params={
            new_exec_conf_uuid: ExtendedRunExecutionParameters(
                uuid=new_exec_conf_uuid,
                name='Foo',
                params=RunExecutionParameters(
                    WorkingDirOpts(source=WorkingDirSource.CurrentDirectory)
                ),
                file_uuid=None
            )
        }
    )

    container.set_executor_configuration_parameters(
        executor_name, 'ext1', RunContext.RegisteredContext, new_params
    )

    new_exec_params = container.get_conf('parameters')[executor_name]

    assert (
        len(
            new_exec_params[('ext1', RunContext.RegisteredContext)]['params']
        ) == 3
    )  # Now we should have two configs in the same context

    # Test teardown functions
    executor_1.on_run_teardown(run)
    executor_2.on_run_teardown(run)

    exec_provider_1.on_run_teardown(run)
    exec_provider_2.on_run_teardown(run)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""External terminal plugin."""

# Standard library imports.
import os
import sys
import tempfile
import platform
from typing import List

# Third-party imports
from packaging.version import parse
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.translations import _
from spyder.plugins.editor.api.run import FileRun, SelectionRun
from spyder.plugins.externalterminal.api import (
    ExtTerminalPyConfiguration, ExtTerminalShConfiguration)
from spyder.plugins.externalterminal.widgets.run_conf import (
    ExternalTerminalPyConfiguration, ExternalTerminalShConfiguration)
from spyder.plugins.mainmenu.api import ApplicationMenus, RunMenuSections
from spyder.plugins.run.api import (
    RunContext, run_execute, RunConfiguration, ExtendedRunExecutionParameters,
    RunResult, RunExecutor)
from spyder.utils import programs
from spyder.utils.misc import get_python_executable


class ExternalTerminal(SpyderPluginV2, RunExecutor):
    """External terminal plugin."""

    NAME = 'external_terminal'
    REQUIRES = [Plugins.Run]
    OPTIONAL = [Plugins.Editor]
    CONF_SECTION = NAME
    CONF_FILE = False

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("External terminal")

    @staticmethod
    def get_description():
        return _("Run scripts in an external system terminal.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('DollarFileIcon')

    def on_initialize(self):
        self.editor_configurations = [
            {
                'origin': self.NAME,
                'extension': 'py',
                'contexts': [{'name': 'File'}]
            }
        ]

        self.executor_configuration = [
            {
                'input_extension': 'py',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalPyConfiguration,
                'requires_cwd': True,
                'priority': 2
            },
        ]

        if os.name == 'nt':
            self.editor_configurations.append({
                'origin': self.NAME,
                'extension': 'bat',
                'contexts': [{'name': 'File'}, {'name': 'Selection'}]
            })

            self.executor_configuration.append({
                'input_extension': 'bat',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalShConfiguration(
                    'cmd.exe', '/K'
                ),
                'requires_cwd': True,
                'priority': 1
            })

            self.executor_configuration.append({
                'input_extension': 'bat',
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalShConfiguration(
                    'cmd.exe', '/K'
                ),
                'requires_cwd': True,
                'priority': 1
            })

            self.editor_configurations.append({
                'origin': self.NAME,
                'extension': 'ps1',
                'contexts': [{'name': 'File'}, {'name': 'Selection'}]
            })

            self.executor_configuration.append({
                'input_extension': 'ps1',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalShConfiguration(
                    'powershell.exe'
                ),
                'requires_cwd': True,
                'priority': 1
            })

            self.executor_configuration.append({
                'input_extension': 'ps1',
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalShConfiguration(
                    'powershell.exe'
                ),
                'requires_cwd': True,
                'priority': 1
            })

        if sys.platform in {'linux', 'darwin'}:
            default_shell = 'bash'
            if sys.platform == 'darwin':
                mac_ver = parse(platform.mac_ver()[0])
                if mac_ver >= parse('10.15.0'):
                    # Catalina changed the default shell to zsh
                    default_shell = 'zsh'

            self.editor_configurations.append({
                'origin': self.NAME,
                'extension': 'sh',
                'contexts': [{'name': 'File'}, {'name': 'Selection'}]
            })

            self.executor_configuration.append({
                'input_extension': 'sh',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalShConfiguration(
                    programs.is_program_installed(default_shell)
                ),
                'requires_cwd': True,
                'priority': 1
            })

            self.executor_configuration.append({
                'input_extension': 'sh',
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': ExternalTerminalShConfiguration(
                    programs.is_program_installed(default_shell)
                ),
                'requires_cwd': True,
                'priority': 1
            })

    @on_plugin_available(plugin=Plugins.Run)
    def on_run_available(self):
        run = self.get_plugin(Plugins.Run)
        run.register_executor_configuration(self, self.executor_configuration)

        run.create_run_in_executor_button(
            RunContext.File,
            self.NAME,
            text=_("Run in external terminal"),
            tip=_("Run in an operating system terminal"),
            icon=self.get_icon(),
            shortcut_context='_',
            register_shortcut=True,
            shortcut_widget_context=Qt.ApplicationShortcut,
            add_to_menu={
                "menu": ApplicationMenus.Run,
                "section": RunMenuSections.RunInExecutors
            },
        )

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        editor = self.get_plugin(Plugins.Editor)
        for conf in self.editor_configurations:
            editor.add_supported_run_configuration(conf)

    @on_plugin_teardown(plugin=Plugins.Run)
    def on_run_teardown(self):
        run = self.get_plugin(Plugins.Run)
        run.deregister_executor_configuration(
            self, self.executor_configuration
        )
        run.destroy_run_in_executor_button(RunContext.File, self.NAME)

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        editor = self.get_plugin(Plugins.Editor)
        for conf in self.editor_configurations:
            editor.remove_supported_run_configuration(conf)

    # ---- Public API
    # -------------------------------------------------------------------------
    def open_external_python_terminal(self, fname, wdir, args, interact, debug,
                                      python_args):
        """Open external terminal."""
        # Running script in an external system terminal
        try:
            if self.get_conf('default', section='main_interpreter'):
                executable = get_python_executable()
            else:
                executable = self.get_conf(
                    'executable',
                    section='main_interpreter'
                )

            pypath = self.get_conf(
                'spyder_pythonpath',
                section='pythonpath_manager',
                default=[]
            )

            programs.run_python_script_in_terminal(
                fname, wdir, args, interact, debug, python_args, executable,
                pypath=pypath
            )
        except NotImplementedError:
            QMessageBox.critical(
                self,
                _("Run"),
                _("Running an external system terminal is not supported on "
                  "platform %s.") % os.name
            )

    @run_execute(extension='py')
    def run_python_files(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:
        exec_params = conf['params']
        cwd_opts = exec_params['working_dir']
        params: ExtTerminalPyConfiguration = exec_params['executor_params']

        run_input: FileRun = input['run_input']
        filename = run_input['path']

        wdir = cwd_opts['path']
        args = params['args']
        interact = params['interact']
        debug = False
        python_args = params['python_args']
        self.open_external_python_terminal(
            filename, wdir, args, interact, debug, python_args
        )

    @run_execute(extension=['sh', 'bat', 'ps1'])
    def run_shell_files(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:
        exec_params = conf['params']
        cwd_opts = exec_params['working_dir']
        params: ExtTerminalShConfiguration = exec_params['executor_params']

        run_input: FileRun = input['run_input']
        filename = run_input['path']
        extension = input['metadata']['input_extension']

        executable = params['interpreter']
        executable_args = params['interpreter_opts']
        script_args = params['script_opts']
        close_after_exec = params['close_after_exec']
        wdir = cwd_opts['path']

        windows_shell = 'cmd.exe'
        no_exit_flag = '/K'
        if extension == 'ps1':
            windows_shell = 'powershell.exe'
            no_exit_flag = '-NoExit'

        if close_after_exec:
            no_exit_flag = ''

        windows_shell = f'{windows_shell} {no_exit_flag}'.strip()

        programs.run_general_file_in_terminal(
            executable, executable_args, filename, script_args, wdir,
            close_after_exec, windows_shell=windows_shell)

    @run_execute(extension=['sh', 'bat', 'ps1'], context=RunContext.Selection)
    def run_shell_selection(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:
        exec_params = conf['params']
        cwd_opts = exec_params['working_dir']
        params: ExtTerminalShConfiguration = exec_params['executor_params']

        metadata = input['metadata']
        run_input: SelectionRun = input['run_input']
        selection = run_input['selection']

        executable = params['interpreter']
        executable_args = params['interpreter_opts']
        script_args = params['script_opts']
        close_after_exec = params['close_after_exec']
        wdir = cwd_opts['path']

        f = tempfile.NamedTemporaryFile(
            'wt', prefix='run_spyder_',
            suffix=f'.{metadata["input_extension"]}',
            dir=programs.get_temp_dir(),
            delete=False)

        with f:
            f.write(selection)
            programs.run_general_file_in_terminal(
                executable, executable_args, f.name, script_args, wdir,
                close_after_exec)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for code formatting using Editor/EditorStack instances."""

# Standard library imports
import random

# Third party imports
from flaky import flaky
import pytest
from qtpy.QtCore import Qt
from qtpy.QtGui import QTextCursor
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder.config.manager import CONF
from spyder.plugins.editor.widgets.codeeditor.tests.conftest import (
    autopep8,
    black,
    get_formatter_values
)


# ---- Tests
@pytest.mark.order(1)
@pytest.mark.parametrize('formatter', [autopep8, black])
def test_closing_document_formatting(
        formatter, completions_editor, qtbot, monkeypatch):
    """Check that auto-formatting works when closing an usaved file."""
    file_path, editorstack, code_editor, completion_plugin = completions_editor
    text, expected = get_formatter_values(formatter, newline='\n')

    # Set formatter
    editorstack.set_format_on_save(True)
    CONF.set(
        'completions',
        ('provider_configuration', 'lsp', 'values', 'formatting'),
        formatter
    )

    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    # Set text in editor
    code_editor.set_text(text)

    # Notify changes
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Perform formatting while closing the file
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        monkeypatch.setattr(QMessageBox, 'exec_',
                            classmethod(lambda *args: QMessageBox.Yes))
        monkeypatch.setattr(editorstack, 'select_savename',
                            lambda *args: str(file_path))
        editorstack.save_dialog_on_tests = True
        editorstack.close_file()

    # Load again formatted file and check content
    code_editor = editorstack.load(str(file_path)).editor

    assert code_editor.get_text_with_eol() == expected


@flaky(max_runs=20)
@pytest.mark.order(1)
@pytest.mark.parametrize('formatter', [autopep8, black])
def test_formatting_on_save(completions_editor, formatter, qtbot):
    """
    Check that auto-formatting on save works as expected and that we restore
    the current line after doing it.

    This includes a regression test for issue spyder-ide/spyder#19958
    """
    file_path, editorstack, code_editor, completion_plugin = completions_editor
    text, expected = get_formatter_values(formatter, newline='\n')

    # Set formatter
    editorstack.set_format_on_save(True)
    CONF.set(
        'completions',
        ('provider_configuration', 'lsp', 'values', 'formatting'),
        formatter
    )

    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    # Set text in editor
    code_editor.set_text(text)

    # Notify changes
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Set a random current line
    current_line = random.randint(0, code_editor.blockCount())
    code_editor.moveCursor(QTextCursor.Start)
    cursor = code_editor.textCursor()
    cursor.movePosition(QTextCursor.Down, QTextCursor.MoveAnchor, current_line)
    code_editor.setTextCursor(cursor)
    qtbot.wait(500)

    # Make a simple change to the file
    code_editor.moveCursor(QTextCursor.EndOfLine)
    qtbot.keyPress(code_editor, Qt.Key_Space)
    current_position = cursor.position()
    qtbot.wait(500)

    # Save the file
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        editorstack.save(force=True)
    qtbot.wait(1000)

    # Check that auto-formatting was applied on save and that we restored the
    # previous line.
    assert code_editor.get_text_with_eol() == expected
    assert code_editor.textCursor().position() == current_position



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Main widget to use in plugins that show content that comes from the IPython
console, such as the Variable Explorer or Plots.
"""

# Third party imports
from qtpy.QtWidgets import QStackedWidget, QVBoxLayout

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.widgets.helperwidgets import PaneEmptyWidget


class ShellConnectMainWidget(PluginMainWidget):
    """
    Main widget to use in a plugin that shows console-specific content.

    Notes
    -----
    * This is composed of a QStackedWidget to stack widgets associated to each
      shell widget in the console and only show one of them at a time.
    * The current widget in the stack will display the content associated to
      the console with focus.
    """
    def __init__(self, *args, set_layout=True, **kwargs):
        super().__init__(*args, **kwargs)

        # Widgets
        self._stack = QStackedWidget(self)
        self._shellwidgets = {}

        if set_layout:
            # Layout
            layout = QVBoxLayout()
            layout.addWidget(self._stack)
            self.setLayout(layout)

    # ---- PluginMainWidget API
    # ------------------------------------------------------------------------
    def current_widget(self):
        """
        Return the current widget in the stack.

        Returns
        -------
        QWidget
            The current widget.
        """
        return self._stack.currentWidget()

    def get_focus_widget(self):
        return self.current_widget()

    # ---- SpyderWidgetMixin API
    # ------------------------------------------------------------------------
    def update_style(self):
        self._stack.setStyleSheet("QStackedWidget {padding: 0px; border: 0px}")

    # ---- Stack accesors
    # ------------------------------------------------------------------------
    def count(self):
        """
        Return the number of widgets in the stack.

        Returns
        -------
        int
            The number of widgets in the stack.
        """
        return self._stack.count()

    def get_widget_for_shellwidget(self, shellwidget):
        """return widget corresponding to shellwidget."""
        shellwidget_id = id(shellwidget)
        if shellwidget_id in self._shellwidgets:
            return self._shellwidgets[shellwidget_id]
        return None

    # ---- Public API
    # ------------------------------------------------------------------------
    def add_shellwidget(self, shellwidget):
        """Create a new widget in the stack and associate it to shellwidget."""
        shellwidget_id = id(shellwidget)
        if shellwidget_id not in self._shellwidgets:
            widget = self.create_new_widget(shellwidget)
            self._stack.addWidget(widget)
            self._shellwidgets[shellwidget_id] = widget

            # Add all actions to new widget for shortcuts to work.
            for __, action in self.get_actions().items():
                if action:
                    widget_actions = widget.actions()
                    if action not in widget_actions:
                        widget.addAction(action)

            self.set_shellwidget(shellwidget)

    def remove_shellwidget(self, shellwidget):
        """Remove widget associated to shellwidget."""
        shellwidget_id = id(shellwidget)
        if shellwidget_id in self._shellwidgets:
            widget = self._shellwidgets.pop(shellwidget_id)

            # If `widget` is an empty pane, we don't need to remove it from the
            # stack (because it's the one we need to show since the console is
            # showing an error) nor try to close it (because it makes no
            # sense).
            if not isinstance(widget, PaneEmptyWidget):
                self._stack.removeWidget(widget)
                self.close_widget(widget)

            self.update_actions()

    def set_shellwidget(self, shellwidget):
        """Set widget associated with shellwidget as the current widget."""
        old_widget = self.current_widget()
        widget = self.get_widget_for_shellwidget(shellwidget)
        if widget is None:
            return

        self._stack.setCurrentWidget(widget)
        self.switch_widget(widget, old_widget)
        self.update_actions()

    def add_errored_shellwidget(self, shellwidget):
        """
        Create a new PaneEmptyWidget in the stack and associate it to
        shellwidget.

        This is necessary to show a meaningful message when switching to
        consoles with dead kernels.
        """
        shellwidget_id = id(shellwidget)

        # This can happen if the kernel started without issues but something is
        # printed to its stderr stream, which we display as an error in the
        # console. In that case, we need to remove the current widget
        # associated to shellwidget and replace it by an empty one.
        if shellwidget_id in self._shellwidgets:
            self._shellwidgets.pop(shellwidget_id)

        widget = PaneEmptyWidget(
            self,
            "console-remote-off" if shellwidget.is_remote() else "console-off",
            _("No connected console"),
            _(
                "The current console has no active kernel, so there is no "
                "content to show here"
            ),
        )
        self._stack.addWidget(widget)
        self._shellwidgets[shellwidget_id] = widget
        self.set_shellwidget(shellwidget)

    def create_new_widget(self, shellwidget):
        """Create a widget to communicate with shellwidget."""
        raise NotImplementedError

    def close_widget(self, widget):
        """Close the widget."""
        raise NotImplementedError

    def switch_widget(self, widget, old_widget):
        """Switch the current widget."""
        raise NotImplementedError

    def refresh(self):
        """Refresh widgets."""
        if self.count():
            widget = self.current_widget()
            widget.refresh()

    def is_current_widget_empty(self):
        """Check if the current widget is a PaneEmptyWidget."""
        return isinstance(self.current_widget(), PaneEmptyWidget)



#!/usr/bin/env python3
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Helper script for installing spyder and external-deps locally in editable mode.
"""

import argparse
from importlib.metadata import PackageNotFoundError, distribution
from json import loads
from logging import Formatter, StreamHandler, getLogger
import os
from pathlib import Path
from subprocess import check_output
import sys

from packaging.requirements import Requirement

# Remove current/script directory from sys.path[0] if added by the Python
# invocation, otherwise Spyder's install status may be incorrectly determined.
SYS_PATH_0 = Path(sys.path[0]).resolve()
if SYS_PATH_0 in (Path(__file__).resolve().parent, Path.cwd()):
    sys.path.pop(0)

DEVPATH = Path(__file__).resolve().parent
DEPS_PATH = DEVPATH / 'external-deps'
BASE_COMMAND = [sys.executable, '-m', 'pip', 'install', '--no-deps']

REPOS = {}
for p in [DEVPATH] + list(DEPS_PATH.iterdir()):
    if (
        p.name.startswith('.')
        or not p.is_dir()
        and not ((p / 'setup.py').exists() or (p / 'pyproject.toml').exists())
    ):
        continue

    try:
        dist = distribution(p.name)
    except PackageNotFoundError:
        dist = None
        editable = None
    else:
        direct_url = dist.read_text('direct_url.json')
        if direct_url:
            editable = (
                loads(direct_url).get('dir_info', {}).get('editable', False)
            )
        else:
            editable = (p == dist._path or p in dist._path.parents)

        # This fixes detecting that PyLSP was installed in editable mode under
        # some scenarios.
        # Fixes spyder-ide/spyder#19712
        if p.name == 'python-lsp-server':
            for f in dist.files:
                if 'editable' in f.name:
                    editable = True
                    break

    REPOS[p.name] = {'repo': p, 'dist': dist, 'editable': editable}

# ---- Setup logger
fmt = Formatter('%(asctime)s [%(levelname)s] [%(name)s] -> %(message)s')
h = StreamHandler()
h.setFormatter(fmt)
logger = getLogger('InstallDevRepos')
logger.addHandler(h)
logger.setLevel('INFO')


def get_python_lsp_version():
    """Get current version to pass it to setuptools-scm."""
    req_file = DEVPATH / 'requirements' / 'main.yml'
    with open(req_file, 'r', encoding='utf-8') as f:
        for line in f:
            if 'python-lsp-server' not in line:
                continue
            line = line.split('-')[-1]
            specifiers = Requirement(line).specifier
            break
        else:
            return "0.0.0"

    for specifier in specifiers:
        if "=" in specifier.operator:
            return specifier.version
    else:
        return "0.0.0"


def install_repo(name, not_editable=False):
    """
    Install a single repo from source located in spyder/external-deps, ignoring
    dependencies, in standard or editable mode.

    Parameters
    ----------
    name : str
        Must be 'spyder' or the distribution name of a repo in
        spyder/external-deps.
    not_editable : bool (False)
        Install repo in standard mode (True) or editable mode (False).
        Editable mode uses pip's `-e` flag.

    """
    try:
        repo_path = REPOS[name]['repo']
    except KeyError:
        logger.warning('Distribution %r not valid. Must be one of %s',
                       name, set(REPOS.keys()))
        return

    install_cmd = BASE_COMMAND.copy()

    # PyLSP requires pretend version
    env = None
    if name == 'python-lsp-server':
        env = {**os.environ}
        env.update(
            {'SETUPTOOLS_SCM_PRETEND_VERSION': get_python_lsp_version()})

    if not_editable:
        mode = 'standard'
    else:
        # Add edit flag to install command
        install_cmd.append('-e')
        mode = 'editable'

    logger.info('Installing %r from source in %s mode.', name, mode)
    install_cmd.append(repo_path.as_posix())
    check_output(install_cmd, env=env)


def main(install=tuple(REPOS.keys()), no_install=tuple(), **kwargs):
    """
    Install all subrepos from source.

    Parameters
    ----------
    install : iterable (spyder and all repos in spyder/external-deps)
        Distribution names of repos to be installed from spyder/external-deps.
    no_install : iterable ()
        Distribution names to exclude from install.
    **kwargs :
        Keyword arguments passed to `install_repo`.

    """
    _install = set(install) - set(no_install)
    for repo in _install:
        install_repo(repo, **kwargs)


if __name__ == '__main__':
    # ---- Parse command line
    parser = argparse.ArgumentParser(
        usage="python install_dev_repos.py [options]")
    parser.add_argument(
        '--install', nargs='+',
        default=REPOS.keys(),
        help="Space-separated list of distribution names to install, e.g. "
             "qtconsole spyder-kernels. If option not provided, then all of "
             "the repos in spyder/external-deps are installed"
    )
    parser.add_argument(
        '--no-install', nargs='+', default=[],
        help="Space-separated list of distribution names to exclude from "
             "install. Default is empty list."
    )
    parser.add_argument(
        '--not-editable', action='store_true', default=False,
        help="Install in standard mode, not editable mode."
    )

    args = parser.parse_args()

    # ---- Install repos locally
    main(**vars(args))



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder Remote Client API.
"""

from __future__ import annotations
import logging
import typing


class KernelInfo(typing.TypedDict):
    id: str
    name: str
    last_activity: str
    execution_state: str
    connections: int
    connection_info: KernelConnectionInfo


class KernelConnectionInfo(typing.TypedDict):
    transport: str
    ip: str
    shell_port: int
    iopub_port: int
    stdin_port: int
    hb_port: int
    control_port: int
    signature_scheme: str
    key: str


class SSHClientOptions(typing.TypedDict):
    host: str
    port: int | None
    username: str
    password: str | None
    client_keys: typing.Sequence[str] | None
    passphrase: str | None
    known_hosts: str | typing.Sequence[str] | None
    config: typing.Sequence[str] | None
    platform: str | None


class ConnectionStatus:
    Inactive = "inactive"
    Connecting = "connecting"
    Active = "active"
    Stopping = "stopping"
    Error = "error"


class ConnectionInfo(typing.TypedDict):
    id: str
    status: ConnectionStatus
    message: str


class RemoteClientLog(typing.TypedDict):
    id: str
    message: str
    level: (
        logging.DEBUG
        | logging.INFO
        | logging.WARNING
        | logging.ERROR
        | logging.CRITICAL
    )
    created: float



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Main interpreter Plugin.
"""

# Standard library imports
import os.path as osp

# Third-party import
from qtpy.QtCore import Signal

# Local imports
from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.plugins.maininterpreter.confpage import MainInterpreterConfigPage
from spyder.plugins.maininterpreter.container import MainInterpreterContainer
from spyder.utils.misc import get_python_executable


class MainInterpreter(SpyderPluginV2):
    """
    Main interpreter Plugin.
    """

    NAME = "main_interpreter"
    REQUIRES = [Plugins.Preferences]
    CONTAINER_CLASS = MainInterpreterContainer
    CONF_WIDGET_CLASS = MainInterpreterConfigPage
    CONF_SECTION = NAME
    CONF_FILE = False
    CAN_BE_DISABLED = False

    # ---- Signals
    # -------------------------------------------------------------------------
    sig_environments_updated = Signal(dict)
    """
    This signal is emitted when the conda, pyenv or custom environments tracked
    by this plugin were updated.

    Parameters
    ----------
    envs: dict
        Environments dictionary in the format given by
        :py:meth:`spyder.utils.envs.get_list_envs`.
    """

    sig_interpreter_changed = Signal(str)
    """
    Signal to report that the main interpreter has changed.

    Parameters
    ----------
    path: str
        Path to the new interpreter.
    """

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Python interpreter")

    @staticmethod
    def get_description():
        return _(
            "Manage the default Python interpreter used to run, analyze and "
            "profile your code in Spyder."
        )

    @classmethod
    def get_icon(cls):
        return cls.create_icon('python')

    def on_initialize(self):
        container = self.get_container()

        # Connect container signals
        container.sig_environments_updated.connect(
            self.sig_environments_updated
        )
        container.sig_interpreter_changed.connect(self.sig_interpreter_changed)

        # Validate that the custom interpreter from the previous session
        # still exists
        if self.get_conf('custom'):
            interpreter = self.get_conf('custom_interpreter')
            if not osp.isfile(interpreter):
                self.set_conf('custom', False)
                self.set_conf('default', True)
                self.set_conf('executable', get_python_executable())

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        # Register conf page
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        # Deregister conf page
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    # ---- Public API
    # -------------------------------------------------------------------------
    def set_custom_interpreter(self, interpreter):
        """Set given interpreter as the current selected one."""
        self.get_container().add_to_custom_interpreters(interpreter)
        self.set_conf("default", False)
        self.set_conf("custom", True)
        self.set_conf("custom_interpreter", interpreter)
        self.set_conf("executable", interpreter)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016 Colin Duquesnoy (QCrash project)
# Copyright (c) 2018- Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/widgets/github/LICENSE.txt for details)
# -----------------------------------------------------------------------------

"""
Tests for the Github backend.

Adapted from tests/test_backends/test_github.py of the
`QCrash Project <https://github.com/ColinDuquesnoy/QCrash>`_.
"""

import os
import sys

import pytest

from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.widgets.github import backend


TOKEN = 'token1234'
GH_OWNER = 'ccordoba12'
GH_REPO = 'spyder'


def get_backend():
    b = backend.GithubBackend(GH_OWNER, GH_REPO)
    b._show_msgbox = False
    return b


def get_backend_bad_repo():
    b = backend.GithubBackend(GH_OWNER, GH_REPO + '1234')
    b._show_msgbox = False
    return b


def get_wrong_user_credentials():
    """
    Monkeypatch GithubBackend.get_user_credentials to force the case where
    invalid credentias were provided
    """
    return dict(token='invalid',
                remember_token=False)


def get_empty_user_credentials():
    """
    Monkeypatch GithubBackend.get_user_credentials to force the case where
    invalid credentias were provided
    """
    return dict(token='',
                remember_token=False)


def get_fake_user_credentials():
    """
    Monkeypatch GithubBackend.get_user_credentials to force the case where
    invalid credentias were provided
    """
    return dict(token=TOKEN,
                remember_token=False)


def test_invalid_credentials():
    b = get_backend()
    b.get_user_credentials = get_wrong_user_credentials
    ret_value = b.send_report('Wrong credentials', 'Wrong credentials')
    assert ret_value is False


def test_empty_credentials():
    b = get_backend()
    b.get_user_credentials = get_empty_user_credentials
    ret_value = b.send_report('Empty credentials', 'Wrong credentials')
    assert ret_value is False


def test_fake_credentials_bad_repo():
    b = get_backend_bad_repo()
    b.get_user_credentials = get_fake_user_credentials
    ret_value = b.send_report('Test suite', 'Test fake credentials')
    assert ret_value is False


def test_get_credentials_from_settings():
    b = get_backend()
    remember_token = b._get_credentials_from_settings()
    assert remember_token is False

    CONF.set('main', 'report_error/remember_token', True)

    remember_token = b._get_credentials_from_settings()
    assert remember_token is True


@pytest.mark.skipif(running_in_ci(), reason="Only works locally")
def test_store_user_credentials():
    b = get_backend()
    b._store_token('token', True)
    credentials = b.get_user_credentials()

    assert credentials['token'] == 'token'
    assert credentials['remember_token'] is True



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


# Standard library imports
import os
import re


root_path = os.path.realpath(os.path.join(os.getcwd(), 'spyder'))

pattern = re.compile("isinstance\(.*,.*str\)")


def test_dont_use_isinstance_str():
    found = False
    for dir_name, _, file_list in os.walk(root_path):
        for fname in file_list:
            if fname.endswith('.py') and fname != 'py3compat.py':
                file = os.path.join(dir_name, fname)

                for i, line in enumerate(open(file)):
                    for match in re.finditer(pattern, line):
                        print("{}\nline:{}, {}".format(file, i + 1, line))
                        found = True

    assert found == False, ("Don't use builtin isinstance() function,"
                            "use spyder.py3compat.is_text_string() instead")



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import os
import re
import codecs

import pytest

root_path = os.path.realpath(os.path.join(os.getcwd(), 'spyder'))


@pytest.mark.parametrize("pattern,exclude_patterns,message", [
    (r"isinstance\(.*,.*str\)", ['py3compat.py', 'example_latin1.py'],
     ("Don't use builtin isinstance() function,"
      "use spyder.py3compat.is_text_string() instead")),

    (r"^[\s\#]*\bprint\(((?!file=).)*\)", ['.*test.*', 'example.py',
                                           'example_latin1.py', 'binaryornot'],
     ("Don't use the print() function; ",
      "for debugging, use logging module instead")),

    (r"^[\s\#]*\bprint\s+(?!>>)((?!#).)*", ['.*test.*', 'example_latin1.py'],
     ("Don't use print statements; ",
      "for debugging, use the logging module instead.")),
])
def test_dont_use(pattern, exclude_patterns, message):
    """
    This test is used for discouraged using of some expressions that could
    introduce errors, and encourage use spyder function instead.

    If you want to skip some line from this test just use:
        # spyder: test-skip
    """
    pattern = re.compile(pattern + r"((?!# spyder: test-skip)\s)*$")

    found = 0
    for dir_name, _, file_list in os.walk(root_path):
        for fname in file_list:
            exclude = any([re.search(ex, fname) for ex in exclude_patterns])
            exclude = exclude or any([re.search(ex, dir_name) for ex in exclude_patterns])

            if fname.endswith('.py') and not exclude:
                file = os.path.join(dir_name, fname)

                with codecs.open(file, encoding="utf-8") as f:
                    for i, line in enumerate(f):
                        for match in re.finditer(pattern, line):
                            print("{}\nline:{}, {}".format(file, i + 1, line))
                            found += 1

    assert found == 0, "{}\n{} errors found".format(message, found)


@pytest.mark.parametrize("pattern", [u"％"])
def test_check_charaters_translation(pattern):
    u"""
    This test is used to prevent the addition of unwanted unicode characters
    in the translations like ％ instead of %.

    """
    found = 0
    for dir_name, _, file_list in os.walk(os.path.join(root_path, 'locale')):
        for fname in file_list:
            if fname.endswith('.po'):
                file = os.path.join(dir_name, fname)

                with codecs.open(file, encoding="utf-8") as f:
                    for i, line in enumerate(f):
                        for match in re.finditer(pattern, line):
                            print(u"{}\nline:{}, {}".format(file, i + 1, line))
                            found += 1

    assert found == 0, u"{}\n{} characters found".format(
        u"Strange characters found in translations", found)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


"""
Client widget for the IPython Console.

This is the widget used on all its tabs.
"""

# Standard library imports.
import functools
import logging
import os
import os.path as osp
from string import Template
import time
import traceback

# Third party imports (qtpy)
from qtpy.QtCore import QUrl, QTimer, Signal, Slot
from qtpy.QtWidgets import QApplication, QVBoxLayout, QWidget

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import (
    get_home_dir, get_module_source_path, get_conf_path)
from spyder.utils.icon_manager import ima
from spyder.utils import sourcecode
from spyder.utils.image_path_manager import get_image_path
from spyder.utils.installers import InstallerIPythonKernelError
from spyder.utils.environ import RemoteEnvDialog
from spyder.utils.palette import SpyderPalette
from spyder.utils.qthelpers import DialogManager
from spyder.plugins.ipythonconsole import SpyderKernelError
from spyder.plugins.ipythonconsole.utils.kernel_handler import (
    KernelConnectionState)
from spyder.plugins.ipythonconsole.widgets import ShellWidget
from spyder.widgets.collectionseditor import CollectionsEditor
from spyder.widgets.mixins import SaveHistoryMixin


# Logging
logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Templates
# -----------------------------------------------------------------------------
# Using the same css file from the Help plugin for now. Maybe
# later it'll be a good idea to create a new one.
PLUGINS_PATH = get_module_source_path('spyder', 'plugins')

CSS_PATH = osp.join(PLUGINS_PATH, 'help', 'utils', 'static', 'css')
TEMPLATES_PATH = osp.join(
    PLUGINS_PATH, 'ipythonconsole', 'assets', 'templates')

BLANK = open(osp.join(TEMPLATES_PATH, 'blank.html')).read()
LOADING = open(osp.join(TEMPLATES_PATH, 'loading.html')).read()
KERNEL_ERROR = open(osp.join(TEMPLATES_PATH, 'kernel_error.html')).read()

try:
    time.monotonic  # time.monotonic new in 3.3
except AttributeError:
    time.monotonic = time.time

# ----------------------------------------------------------------------------
# Client widget
# ----------------------------------------------------------------------------
class ClientWidget(QWidget, SaveHistoryMixin, SpyderWidgetMixin):
    """
    Client widget for the IPython Console

    This widget is necessary to handle the interaction between the
    plugin and each shell widget.
    """

    sig_append_to_history_requested = Signal(str, str)
    sig_execution_state_changed = Signal()
    sig_time_label = Signal(str)

    # Signals for remote kernels
    sig_shutdown_kernel_requested = Signal(str, str)
    sig_interrupt_kernel_requested = Signal(str, str)
    sig_restart_kernel_requested = Signal()
    sig_kernel_died = Signal()

    CONF_SECTION = 'ipython_console'
    SEPARATOR = '{0}## ---({1})---'.format(os.linesep*2, time.ctime())
    INITHISTORY = ['# -*- coding: utf-8 -*-',
                   '# *** Spyder Python Console History Log ***', ]

    def __init__(
        self,
        parent,
        id_,
        config_options,
        additional_options,
        menu_actions=None,
        given_name=None,
        give_focus=True,
        options_button=None,
        handlers=None,
        initial_cwd=None,
        forcing_custom_interpreter=False,
        special_kernel=None,
        server_id=None,
        can_close=True,
    ):
        super(ClientWidget, self).__init__(parent)
        SaveHistoryMixin.__init__(self, get_conf_path('history.py'))

        # --- Init attrs
        self.container = parent
        self.id_ = id_
        self.menu_actions = menu_actions
        self.given_name = given_name
        self.initial_cwd = initial_cwd
        self.forcing_custom_interpreter = forcing_custom_interpreter
        self.server_id = server_id
        self.can_close = can_close

        # --- Other attrs
        self.kernel_handler = None
        self.hostname = None
        self.show_elapsed_time = self.get_conf('show_elapsed_time')
        self.reset_warning = self.get_conf('show_reset_namespace_warning')
        self.options_button = options_button
        self.history = []
        self.allow_rename = True
        self.error_text = None
        self.give_focus = give_focus
        self.kernel_id = None
        self.__on_close = lambda: None

        css_path = self.get_conf('css_path', section='appearance')
        if css_path is None:
            self.css_path = CSS_PATH
        else:
            self.css_path = css_path

        # --- Widgets
        self.shellwidget = ShellWidget(
            config=config_options,
            ipyclient=self,
            additional_options=additional_options,
            handlers=handlers,
            local_kernel=True,
            special_kernel=special_kernel,
            server_id=server_id,
        )
        self.infowidget = self.container.infowidget
        self.blank_page = self._create_blank_page()
        self.loading_page = self._create_loading_page()
        # To keep a reference to the page to be displayed
        # in infowidget
        self.info_page = None

        # Elapsed time
        self.t0 = time.monotonic()
        self.timer = QTimer(self)

        # --- Layout
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.shellwidget)
        if self.infowidget is not None:
            self.layout.addWidget(self.infowidget)
        self.setLayout(self.layout)

        # --- Exit function
        self.exit_callback = lambda: self.container.close_client(client=self)

        # --- Dialog manager
        self.dialog_manager = DialogManager()

    # ---- Private methods
    # -------------------------------------------------------------------------
    def _when_kernel_is_ready(self):
        """
        Configuration after the prompt is shown.

        Notes
        -----
        This is not called on restart. For kernel setup you need to use
        ShellWidget.handle_kernel_is_ready.
        """
        if self.kernel_handler.connection_state not in [
                KernelConnectionState.SpyderKernelReady,
                KernelConnectionState.IpykernelReady]:
            # The kernel is not ready
            return

        self.kernel_handler.sig_kernel_is_ready.disconnect(
            self._when_kernel_is_ready)

        # To hide the loading page
        self._hide_loading_page()

        # Set the initial current working directory in the kernel
        self._set_initial_cwd_in_kernel()

        # Notes:
        # 1. It's necessary to do this at this point to avoid giving focus to
        #    _control at startup.
        # 2. The try except is needed to avoid some errors in our tests.
        try:
            self._connect_control_signals()
        except RuntimeError:
            pass

        if self.give_focus:
            self.shellwidget._control.setFocus()

    def _create_loading_page(self):
        """Create html page to show while the kernel is starting"""
        loading_template = Template(LOADING)
        loading_img = get_image_path('loading_sprites')
        if os.name == 'nt':
            loading_img = loading_img.replace('\\', '/')
        message = _("Connecting to kernel...")
        page = loading_template.substitute(css_path=self.css_path,
                                           loading_img=loading_img,
                                           message=message)
        return page

    def _create_blank_page(self):
        """Create html page to show while the kernel is starting"""
        loading_template = Template(BLANK)
        page = loading_template.substitute(css_path=self.css_path)
        return page

    def _show_loading_page(self):
        """Show animation while the kernel is loading."""
        if self.infowidget is not None:
            self.shellwidget.hide()
            self.infowidget.show()
            self.info_page = self.loading_page
            self.set_info_page()

    def _hide_loading_page(self):
        """Hide animation shown while the kernel is loading."""
        if self.infowidget is not None:
            self.infowidget.hide()
            self.info_page = self.blank_page
            self.set_info_page()
        self.shellwidget.show()

    def _show_special_console_error(self, missing_dependency):
        if missing_dependency is not None:
            error_message = _(
                "Your Python environment or installation doesn't have the "
                "<tt>{missing_dependency}</tt> module installed or it "
                "occurred a problem importing it. Due to that, it is not "
                "possible for Spyder to create this special console for "
                "you."
            ).format(missing_dependency=missing_dependency)

            self.show_kernel_error(error_message)

    def _connect_control_signals(self):
        """Connect signals of control widgets."""
        control = self.shellwidget._control
        page_control = self.shellwidget._page_control

        control.sig_focus_changed.connect(
            self.container.sig_focus_changed)
        page_control.sig_focus_changed.connect(
            self.container.sig_focus_changed)
        control.sig_visibility_changed.connect(
            self.container.refresh_container)
        page_control.sig_visibility_changed.connect(
            self.container.refresh_container)
        page_control.sig_show_find_widget_requested.connect(
            self.container.find_widget.show)

    def _set_initial_cwd_in_kernel(self):
        """Set the initial cwd in the kernel."""
        logger.debug("Setting initial working directory in the kernel")
        cwd_path = get_home_dir()
        project_path = self.container.get_active_project_path()
        emit_cwd_change = True

        # This is for the first client
        if self.id_['int_id'] == '1':
            if self.get_conf(
                'startup/use_project_or_home_directory',
                section='workingdir'
            ):
                cwd_path = get_home_dir()
                if project_path is not None:
                    cwd_path = project_path
            elif self.get_conf(
                'startup/use_fixed_directory',
                section='workingdir'
            ):
                cwd_path = self.get_conf(
                    'startup/fixed_directory',
                    default=get_home_dir(),
                    section='workingdir'
                )
        else:
            # For new clients
            if self.initial_cwd is not None:
                cwd_path = self.initial_cwd
            elif self.get_conf(
                'console/use_project_or_home_directory',
                section='workingdir'
            ):
                cwd_path = get_home_dir()
                if project_path is not None:
                    cwd_path = project_path
            elif self.get_conf('console/use_cwd', section='workingdir'):
                cwd_path = self.container.get_working_directory()
                emit_cwd_change = False
            elif self.get_conf(
                'console/use_fixed_directory',
                section='workingdir'
            ):
                cwd_path = self.get_conf(
                    'console/fixed_directory',
                    default=get_home_dir(),
                    section='workingdir'
                )

        if osp.isdir(cwd_path):
            self.shellwidget.set_cwd(cwd_path, emit_cwd_change=emit_cwd_change)

    # ---- Public API
    # -------------------------------------------------------------------------
    @property
    def connection_file(self):
        if self.kernel_handler is None:
            return None
        return self.kernel_handler.connection_file

    def connect_kernel(self, kernel_handler, first_connect=True):
        """Connect kernel to client using our handler."""
        self.kernel_handler = kernel_handler

        # Connect standard streams.
        kernel_handler.sig_stderr.connect(self.print_stderr)
        kernel_handler.sig_stdout.connect(self.print_stdout)
        kernel_handler.sig_fault.connect(self.print_fault)
        kernel_handler.sig_kernel_is_ready.connect(
            self._when_kernel_is_ready)

        if self.is_remote():
            self._hide_loading_page()
        else:
            self._show_loading_page()

        # Actually do the connection
        self.shellwidget.connect_kernel(kernel_handler, first_connect)

    def disconnect_kernel(self, shutdown_kernel):
        """Disconnect from current kernel."""
        kernel_handler = getattr(self, "kernel_handler", None)
        if not kernel_handler:
            return

        kernel_handler.sig_stderr.disconnect(self.print_stderr)
        kernel_handler.sig_stdout.disconnect(self.print_stdout)
        kernel_handler.sig_fault.disconnect(self.print_fault)

        self.shellwidget.disconnect_kernel(shutdown_kernel)
        self.kernel_handler = None

    @Slot(str)
    def print_stderr(self, stderr):
        """Print stderr written in PIPE."""
        if not stderr:
            return

        if self.is_benign_error(stderr):
            return

        if self.shellwidget.isHidden():
            error_text = '<tt>%s</tt>' % stderr
            # Avoid printing the same thing again
            if self.error_text != error_text:
                if self.error_text:
                    # Append to error text
                    error_text = self.error_text + error_text
                self.show_kernel_error(error_text)

        if self.shellwidget._starting:
            self.shellwidget.banner = (
                stderr + '\n' + self.shellwidget.banner)
        else:
            self.shellwidget._append_plain_text(
                stderr, before_prompt=True)

    @Slot(str)
    def print_stdout(self, stdout):
        """Print stdout written in PIPE."""
        if not stdout:
            return

        if self.shellwidget._starting:
            self.shellwidget.banner = (
                stdout + '\n' + self.shellwidget.banner)
        else:
            self.shellwidget._append_plain_text(
                stdout, before_prompt=True)

    def connect_shellwidget_signals(self):
        """Configure shellwidget after kernel is connected."""
        # Set exit callback
        self.shellwidget.exit_requested.connect(self.exit_callback)

        # To save history
        self.shellwidget.executing.connect(self.add_to_history)

        # For Mayavi to run correctly
        self.shellwidget.executing.connect(
            self.shellwidget.set_backend_for_mayavi)

        # To update history after execution
        self.shellwidget.executed.connect(self.update_history)

        # To enable the stop button when executing a process
        self.shellwidget.executing.connect(
            self.sig_execution_state_changed)

        # To disable the stop button after execution stopped
        self.shellwidget.executed.connect(
            self.sig_execution_state_changed)

        # To correctly change Matplotlib backend interactively
        self.shellwidget.executing.connect(
            self.shellwidget.change_mpl_backend)

        # To show env and sys.path contents
        self.shellwidget.sig_show_syspath.connect(self.show_syspath)
        self.shellwidget.sig_show_env.connect(self.show_env)

    def add_to_history(self, command):
        """Add command to history"""
        if self.shellwidget.is_debugging():
            return
        return super(ClientWidget, self).add_to_history(command)

    def is_client_executing(self):
        return (self.shellwidget._executing or
                self.shellwidget.is_waiting_pdb_input())

    @Slot()
    def stop_button_click_handler(self):
        """Method to handle what to do when the stop button is pressed"""
        # Interrupt computations or stop debugging
        if not self.shellwidget.is_waiting_pdb_input():
            self.interrupt_kernel()
        else:
            self.shellwidget.pdb_execute_command('exit')

    def show_kernel_error(self, error):
        """Show kernel initialization errors in infowidget."""
        if isinstance(error, Exception):
            if isinstance(error, SpyderKernelError):
                error = error.args[0]
            else:
                error = _("The error is:<br><br>"
                          "<tt>{}</tt>").format(traceback.format_exc())
        self.error_text = error

        if self.is_benign_error(error):
            return

        InstallerIPythonKernelError(error)

        # Replace end of line chars with <br>
        eol = sourcecode.get_eol_chars(error)
        if eol:
            error = error.replace(eol, '<br>')

        # Don't break lines in hyphens
        # From https://stackoverflow.com/q/7691569/438386
        error = error.replace('-', '&#8209')

        # Create error page
        message = _("An error occurred while starting the kernel")
        kernel_error_template = Template(KERNEL_ERROR)
        self.info_page = kernel_error_template.substitute(
            css_path=self.css_path,
            message=message,
            error=error)

        # Show error
        if self.infowidget is not None:
            self.set_info_page()
            self.shellwidget.hide()
            self.infowidget.show()

        # Inform other plugins that the shell failed to start
        self.shellwidget.sig_shellwidget_errored.emit(self.shellwidget)

        # Stop shellwidget
        self.shellwidget.shutdown()

    def show_kernel_connection_error(self):
        self.show_kernel_error(
            _(
                "It was not possible to connect to the kernel associated to "
                "this console. If you are trying to connect to an existing "
                "kernel, check that the connection file you selected actually "
                "corresponds to the kernel you want to connect to."
            )
        )

    def is_benign_error(self, error):
        """Decide if an error is benign in order to filter it."""
        benign_errors = [
            # Error when switching from the Qt5 backend to the Tk one.
            # See spyder-ide/spyder#17488
            "KeyboardInterrupt caught in kernel",
            "QSocketNotifier: Multiple socket notifiers for same socket",
            # Error when switching from the Tk backend to the Qt5 one.
            # See spyder-ide/spyder#17488
            "Tcl_AsyncDelete async handler deleted by the wrong thread",
            "error in background error handler:",
            "    while executing",
            '"::tcl::Bgerror',
            # Avoid showing this warning because it was up to the user to
            # disable secure writes.
            "WARNING: Insecure writes have been enabled via environment",
            # Old error
            "No such comm",
            # PYDEVD debug warning message. See spyder-ide/spyder#18908
            "Note: Debugging will proceed. "
            "Set PYDEVD_DISABLE_FILE_VALIDATION=1 to disable this validation.",
            # Argument not expected error. See spyder-ide/spyder#19298
            "The following argument was not expected",
            # Avoid showing error for kernel restarts after kernel dies when
            # using an external interpreter
            "conda.cli.main_run",
            # Warning when debugpy is not available because it's an optional
            # dependency of IPykernel.
            # See spyder-ide/spyder#21900
            "debugpy_stream undefined, debugging will not be enabled",
            # Harmless warning from OpenCL on Windows.
            # See spyder-ide/spyder#22551
            "The system cannot find the path specified",
            # UNC paths, see spyder-ide/spyder#23726
            " UNC ",
        ]

        return any([err in error for err in benign_errors])

    def get_name(self):
        """Return client name"""
        if self.given_name is None:
            # Name according to host
            if self.hostname is None:
                name = _("Console")
            else:
                name = self.hostname
            # Adding id to name
            client_id = self.id_['int_id'] + u'/' + self.id_['str_id']
            name = name + u' ' + client_id
        elif (self.given_name in ["Pylab", "SymPy", "Cython"] or
              self.forcing_custom_interpreter):
            client_id = self.id_['int_id'] + u'/' + self.id_['str_id']
            name = self.given_name + u' ' + client_id
        else:
            name = self.given_name + u'/' + self.id_['str_id']
        return name

    def get_control(self):
        """Return the text widget (or similar) to give focus to"""
        # page_control is the widget used for paging
        page_control = self.shellwidget._page_control
        if page_control and page_control.isVisible():
            return page_control
        else:
            return self.shellwidget._control

    def set_font(self, font):
        """Set IPython widget's font"""
        self.shellwidget._control.setFont(font)
        self.shellwidget.font = font

    def set_color_scheme(self, color_scheme, reset=True):
        """Set IPython color scheme."""
        # Needed to handle not initialized kernel_client
        # See spyder-ide/spyder#6996.
        try:
            self.shellwidget.set_color_scheme(color_scheme, reset)
        except AttributeError:
            pass

    def close_client(self, is_last_client, close_console=False):
        """Close the client."""
        self.__on_close = lambda: None
        debugging = False

        # Needed to handle a RuntimeError. See spyder-ide/spyder#5568.
        try:
            # This is required after spyder-ide/spyder#21788 to prevent freezes
            # when closing Spyder. That happens not only when a console is in
            # debugging mode before closing, but also when a kernel restart is
            # requested while debugging.
            if self.shellwidget.is_debugging():
                debugging = True
                self.__on_close = functools.partial(
                    self.finish_close,
                    is_last_client,
                    close_console,
                    debugging
                )
                self.shellwidget.sig_prompt_ready.connect(self.__on_close)
                self.shellwidget.stop_debugging()
            else:
                self.interrupt_kernel()
        except RuntimeError:
            pass

        if not debugging:
            self.finish_close(is_last_client, close_console, debugging)

    def finish_close(self, is_last_client, close_console, debugging):
        """Actions to take to finish closing the client."""
        # Disconnect timer needed to update elapsed time and this slot in case
        # it was connected.
        try:
            self.shellwidget.sig_prompt_ready.disconnect(self.__on_close)
            self.timer.timeout.disconnect(self.show_time)
        except (RuntimeError, TypeError):
            pass

        # This is a hack to prevent segfaults when closing Spyder and the
        # client was debugging before doing it.
        # It's a side effect of spyder-ide/spyder#21788
        if debugging and close_console:
            for __ in range(3):
                time.sleep(0.08)
                QApplication.processEvents()

        self.shutdown(is_last_client, close_console=close_console)

        # Prevent errors in our tests
        try:
            self.close()
            self.setParent(None)
        except RuntimeError:
            pass

    def shutdown(self, is_last_client, close_console=False):
        """Shutdown connection and kernel if needed."""
        self.dialog_manager.close_all()
        shutdown_kernel = (
            is_last_client
            and (not self.shellwidget.is_external_kernel or self.is_remote())
            and not self.error_text
        )

        if self.is_remote() and shutdown_kernel and not close_console:
            # This signal allows to shutdown a remote kernel when a client is
            # closed. And we don't emit it when the console is being closed
            # because it's not necessary in that case.
            self.sig_shutdown_kernel_requested.emit(
                self.server_id, self.kernel_id
            )

        self.shellwidget.shutdown(shutdown_kernel)

    def interrupt_kernel(self):
        """Interrupt the associanted Spyder kernel if it's running"""
        # Needed to prevent a crash when a kernel is not running.
        # See spyder-ide/spyder#6299.
        try:
            self.shellwidget.request_interrupt_kernel()
        except RuntimeError:
            pass

    def replace_kernel(self, kernel_handler, shutdown_kernel, clear=True):
        """
        Replace kernel by disconnecting from the current one and connecting to
        another kernel, which is equivalent to a restart.
        """
        # Connect kernel to client
        self.disconnect_kernel(shutdown_kernel)
        self.connect_kernel(kernel_handler, first_connect=False)

        # Reset shellwidget and print restart message
        self.shellwidget.reset(clear=clear)
        self.shellwidget._kernel_restarted_message(died=False)

    def is_kernel_active(self):
        """Check if the kernel is active."""
        return (
            self.kernel_handler is not None
            and self.kernel_handler.connection_state
            in [
                KernelConnectionState.SpyderKernelReady,
                KernelConnectionState.IpykernelReady,
            ]
        )

    def print_fault(self, fault):
        """Print fault text."""
        self.shellwidget._append_plain_text('\n' + fault, before_prompt=True)

    @Slot()
    def enter_array_inline(self):
        """Enter and show the array builder on inline mode."""
        self.shellwidget._control.enter_array_inline()

    @Slot()
    def enter_array_table(self):
        """Enter and show the array builder on table."""
        self.shellwidget._control.enter_array_table()

    @Slot()
    def inspect_object(self):
        """Show how to inspect an object with our Help plugin"""
        self.shellwidget._control.inspect_current_object()

    @Slot()
    def clear_line(self):
        """Clear a console line"""
        self.shellwidget._keyboard_quit()

    @Slot()
    def clear_console(self):
        """Clear the whole console"""
        self.shellwidget.clear_console()

    @Slot()
    def reset_namespace(self):
        """Resets the namespace by removing all names defined by the user"""
        self.shellwidget.reset_namespace(warning=self.reset_warning,
                                         message=True)

    def update_history(self):
        self.history = self.shellwidget._history

    @Slot(object)
    def show_syspath(self, syspath):
        """Show sys.path contents."""
        if syspath is not None:
            editor = CollectionsEditor(self)
            editor.setup(syspath, title="sys.path contents", readonly=True,
                         icon=ima.icon('syspath'))
            self.dialog_manager.show(editor)
        else:
            return

    @Slot(object)
    def show_env(self, env):
        """Show environment variables."""
        self.dialog_manager.show(RemoteEnvDialog(env, parent=self))

    def show_time(self, end=False):
        """Text to show in time_label."""

        elapsed_time = time.monotonic() - self.t0
        # System time changed to past date, so reset start.
        if elapsed_time < 0:
            self.t0 = time.monotonic()
            elapsed_time = 0
        if elapsed_time > 24 * 3600:  # More than a day...!
            fmt = "%d %H:%M:%S"
        else:
            fmt = "%H:%M:%S"
        if end:
            color = SpyderPalette.COLOR_TEXT_3
        else:
            color = SpyderPalette.COLOR_ACCENT_4
        text = "<span style=\'color: %s\'><b>%s" \
               "</b></span>" % (color,
                                time.strftime(fmt, time.gmtime(elapsed_time)))
        if self.show_elapsed_time:
            self.sig_time_label.emit(text)
        else:
            self.sig_time_label.emit("")

    @Slot(bool)
    def set_show_elapsed_time(self, state):
        """Slot to show/hide elapsed time label."""
        self.show_elapsed_time = state

    def set_info_page(self):
        """Set current info_page."""
        if self.infowidget is not None and self.info_page is not None:
            self.infowidget.setHtml(
                self.info_page,
                QUrl.fromLocalFile(self.css_path)
            )
            self.sig_execution_state_changed.emit()

    # ---- For remote clients
    # -------------------------------------------------------------------------
    def is_remote(self):
        """Check if this client is connected to a remote server."""
        return self.server_id is not None

    def handle_remote_kernel_restarted(self, clear=True):
        """Handle restarts for remote kernels."""
        # Reset shellwidget and print restart message
        self.shellwidget.reset(clear=clear)

    def show_restarting_message(self, died=False):
        self.shellwidget._kernel_restarted_message(died=died)

    def remote_kernel_restarted_failure_message(
        self, error=None, shutdown=False
    ):
        """Show message when the kernel failed to be restarted."""

        msg = _("It was not possible to restart the kernel")

        if error is None:
            error_html = f"<br>{msg}<br>"
        else:
            if isinstance(error, SpyderKernelError):
                error = error.args[0]
            elif isinstance(error, Exception):
                error = _("The error is:<br><br>" "<tt>{}</tt>").format(
                    traceback.format_exc()
                )

            # Replace end of line chars with <br>
            eol = sourcecode.get_eol_chars(error)
            if eol:
                error = error.replace(eol, '<br>')

            # Don't break lines in hyphens
            # From https://stackoverflow.com/q/7691569/438386
            error = error.replace('-', '&#8209')

            # Create error page
            kernel_error_template = Template(KERNEL_ERROR)
            error_html = kernel_error_template.substitute(
                css_path=self.css_path,
                message=msg,
                error=error)

        self.shellwidget._append_html(error_html, before_prompt=False)
        self.shellwidget.insert_horizontal_ruler()

        if shutdown:
            self.shutdown(is_last_client=False, close_console=False)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for editor codeanalysis warnings."""

# Stdlib imports
import sys

# Third party imports
from flaky import flaky
import pytest
from qtpy.QtCore import Qt

# Local imports
from spyder.config.manager import CONF


TEXT = ("def some_function():\n"  # D100, D103: Missing docstring
        "    \n"  # W293 trailing spaces
        "    a = 1 # a comment\n"  # E261 two spaces before inline comment
        "\n"
        "    a += s\n"  # Undefined variable s
        "    return a\n"
        "undefined_function()")  # Undefined name 'undefined_function'


@pytest.fixture
def completions_codeeditor_linting(request, qtbot, completions_codeeditor):
    editor, completion_plugin = completions_codeeditor
    CONF.set('completions',
             ('provider_configuration', 'lsp', 'values', 'pydocstyle'),
             True)

    CONF.set('completions',
             ('provider_configuration', 'lsp', 'values', 'pycodestyle'),
             True)

    # After this call the manager needs to be reinitialized
    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    def teardown():
        CONF.set('completions',
                 ('provider_configuration', 'lsp', 'values', 'pydocstyle'),
                 False)

        CONF.set('completions',
                 ('provider_configuration', 'lsp', 'values', 'pycodestyle'),
                 False)

        # After this call the manager needs to be reinitialized
        completion_plugin.after_configuration_update([])
        qtbot.wait(2000)

    request.addfinalizer(teardown)

    return editor, completion_plugin


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_ignore_warnings(qtbot, completions_codeeditor_linting):
    """Test that the editor is ignoring some warnings."""
    editor, completion_plugin = completions_codeeditor_linting

    # Set text in editor
    editor.set_text(TEXT)

    CONF.set('completions',
             ('provider_configuration', 'lsp', 'values', 'pydocstyle/ignore'),
             'D100')

    CONF.set('completions',
             ('provider_configuration', 'lsp', 'values', 'pycodestyle/ignore'),
             'E261')

    # After this call the manager needs to be reinitialized
    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    # Notify changes
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Get current warnings
    qtbot.wait(2000)
    warnings = editor.get_current_warnings()

    expected = [['D103: Missing docstring in public function', 1],
                ['W293 blank line contains whitespace', 2],
                ["undefined name 's'", 5],
                ["undefined name 'undefined_function'", 7],
                ["W292 no newline at end of file", 7],
                ["""E305 expected 2 blank lines after class or """
                 """function definition, found 0""", 7]]

    CONF.set('completions',
             ('provider_configuration', 'lsp', 'values', 'pydocstyle/ignore'),
             '')

    CONF.set('completions',
             ('provider_configuration', 'lsp', 'values', 'pycodestyle/ignore'),
             '')

    completion_plugin.after_configuration_update([])
    qtbot.wait(2000)

    assert warnings == expected


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_adding_warnings(qtbot, completions_codeeditor_linting):
    """Test that warnings are saved in the editor blocks."""
    editor, _ = completions_codeeditor_linting

    # Set text in editor
    editor.set_text(TEXT)

    # Notify changes
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    qtbot.wait(2000)
    block = editor.textCursor().block()
    line_count = editor.document().blockCount()

    warnings = []
    for i in range(line_count):
        data = block.userData()
        if data:
            for analysis in data.code_analysis:
                warnings.append((i+1, analysis[-1]))
        block = block.next()

    expected_warnings = {1: ['D100', 'D103'],
                         2: ['W293'],
                         3: ['E261'],
                         5: ['undefined name'],
                         7: ['undefined name', 'W292', 'E305']}
    for i, warning in warnings:
        assert any([expected in warning for expected in expected_warnings[i]])


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_move_warnings(qtbot, completions_codeeditor_linting):
    """Test that moving to next/previous warnings is working."""
    editor, _ = completions_codeeditor_linting

    # Set text in editor
    editor.set_text(TEXT)

    # Notify changes
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Wait for linting info to arrive
    qtbot.wait(2000)

    # Move between warnings
    editor.go_to_next_warning()
    assert 2 == editor.get_cursor_line_number()

    editor.go_to_next_warning()
    assert 3 == editor.get_cursor_line_number()

    editor.go_to_previous_warning()
    assert 2 == editor.get_cursor_line_number()

    # Test cycling behaviour
    editor.go_to_line(7)
    editor.go_to_next_warning()
    assert 1 == editor.get_cursor_line_number()

    editor.go_to_previous_warning()
    assert 7 == editor.get_cursor_line_number()


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_get_warnings(qtbot, completions_codeeditor_linting):
    """Test that the editor is returning the right list of warnings."""
    editor, _ = completions_codeeditor_linting

    # Set text in editor
    editor.set_text(TEXT)

    # Notify changes
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Wait for linting info to arrive
    qtbot.wait(2000)

    # Get current warnings
    warnings = editor.get_current_warnings()

    expected = [['D100: Missing docstring in public module', 1],
                ['D103: Missing docstring in public function', 1],
                ['W293 blank line contains whitespace', 2],
                ['E261 at least two spaces before inline comment', 3],
                ["undefined name 's'", 5],
                ["undefined name 'undefined_function'", 7],
                ["W292 no newline at end of file", 7],
                ["""E305 expected 2 blank lines after class or """
                 """function definition, found 0""", 7]]

    assert warnings == expected


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_update_warnings_after_delete_line(qtbot, completions_codeeditor_linting):
    """
    Test that code style warnings are correctly updated after deleting a line
    in the Editor.

    Regression test for spyder-ide/spyder#9299.
    """
    editor, _ = completions_codeeditor_linting
    editor.set_text(TEXT)

    # Notify changes.
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Wait for linting info to arrive
    qtbot.wait(2000)

    # Delete the blank line that is causing the W293 warning on line 2.
    editor.go_to_line(2)
    editor.delete_line()

    # Wait for the lsp_response_signal.
    qtbot.waitSignal(editor.completions_response_signal, timeout=30000)

    # Assert that the W293 warning is gone.
    expected = [['D100: Missing docstring in public module', 1],
                ['D103: Missing docstring in public function', 1],
                ['E261 at least two spaces before inline comment', 2],
                ["undefined name 's'", 4],
                ["undefined name 'undefined_function'", 6],
                ["W292 no newline at end of file", 6],
                ["""E305 expected 2 blank lines after class or """
                 """function definition, found 0""", 6]]
    assert editor.get_current_warnings() == expected


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_update_warnings_after_closequotes(qtbot, completions_codeeditor_linting):
    """
    Test that code errors are correctly updated after activating closequotes
    in the Editor.

    Regression test for spyder-ide/spyder#9323.
    """
    editor, _ = completions_codeeditor_linting
    editor.textCursor().insertText("print('test)\n")

    if sys.version_info >= (3, 12):
        expected = [
            ['unterminated string literal (detected at line 1)', 1],
            ['E901 TokenError: unterminated string literal (detected at line 1)', 1]
        ]
    elif sys.version_info >= (3, 10):
        expected = [['unterminated string literal (detected at line 1)', 1]]
    else:
        expected = [['EOL while scanning string literal', 1]]

    # Notify changes.
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Wait for linting info to arrive
    qtbot.wait(2000)

    assert editor.get_current_warnings() == expected

    # Wait for the lsp_response_signal.
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        # Add a single quote to fix the error
        editor.move_cursor(-2)
        qtbot.keyPress(editor, Qt.Key_Apostrophe)
        assert editor.toPlainText() == "print('test')\n"

    # Assert that the error is gone.
    qtbot.wait(2000)
    expected = [['D100: Missing docstring in public module', 1]]
    assert editor.get_current_warnings() == expected


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_update_warnings_after_closebrackets(qtbot, completions_codeeditor_linting):
    """
    Test that code errors are correctly updated after activating closebrackets
    in the Editor.

    Regression test for spyder-ide/spyder#9323.
    """
    editor, _ = completions_codeeditor_linting
    editor.textCursor().insertText("print('test'\n")

    if sys.version_info >= (3, 12):
        expected = [
            ["'(' was never closed", 1],
            ['E901 TokenError: unexpected EOF in multi-line statement', 1]
        ]
    elif sys.version_info >= (3, 10):
        expected = [
            ["'(' was never closed", 1],
            ['E901 TokenError: EOF in multi-line statement', 2]
        ]
    else:
        expected = [
            ['unexpected EOF while parsing', 1],
            ['E901 TokenError: EOF in multi-line statement', 2]
        ]

    # Notify changes.
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Wait for linting info to arrive
    qtbot.wait(2000)

    assert editor.get_current_warnings() == expected

    # Wait for the lsp_response_signal.
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        # Add a bracket to fix the error
        editor.move_cursor(-1)
        qtbot.keyPress(editor, Qt.Key_ParenRight)
        assert editor.toPlainText() == "print('test')\n"

    # Assert that the error is gone.
    qtbot.wait(2000)
    expected = [['D100: Missing docstring in public module', 1]]
    assert editor.get_current_warnings() == expected


@pytest.mark.order(2)
@flaky(max_runs=5)
@pytest.mark.parametrize(
    'ignore_comment', ['#noqa', '# NOQA', '# analysis:ignore', '# no-work']
)
def test_ignore_warnings_with_comments(
    qtbot, completions_codeeditor_linting, ignore_comment
):
    """
    Test that ignoring linting messages with inline comments works as
    expected.
    """
    editor, _ = completions_codeeditor_linting
    editor.textCursor().insertText(
        f"foo {ignore_comment}\n"
        "bar\n"
    )

    if ignore_comment == '# no-work':
        expected = [
            ["undefined name 'foo'", 1],
            ['D100: Missing docstring in public module', 1],
            ['E261 at least two spaces before inline comment', 1],
            ["undefined name 'bar'", 2]
        ]
    else:
        expected = [["undefined name 'bar'", 2]]

    # Notify changes.
    with qtbot.waitSignal(editor.completions_response_signal, timeout=30000):
        editor.document_did_change()

    # Wait for linting info to arrive
    qtbot.wait(2000)

    assert editor.get_current_warnings() == expected



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import json


def parse_notebook_cells(notebook_path):
    with open(notebook_path) as f:
        notebook_data = json.load(f)

    cells = []
    for cell in notebook_data["cells"]:
        if cell["cell_type"] == "code":
            source = "".join(cell["source"])
            outputs = []
            for output in cell["outputs"]:
                if output["output_type"] == "stream":
                    outputs.append("".join(output["text"]))
                elif output["output_type"] == "execute_result":
                    outputs.append("".join(output["data"]["text/plain"]))
            result = "\n".join(outputs)
            cells.append((source, result))

    return cells


def render_notebook(cells):
    notebook_template = {
        "cells": [],
        "nbformat": 4,
        "nbformat_minor": 4,
        "metadata": {},
    }

    for i, (code, result) in enumerate(cells, start=1):
        notebook_template["cells"].append(
            {
                "cell_type": "code",
                "execution_count": i,
                "metadata": {},
                "outputs": [
                    {
                        "data": {"text/plain": result},
                        "execution_count": i,
                        "metadata": {},
                        "output_type": "execute_result",
                    }
                ],
                "source": code,
            }
        )

    return notebook_template


TEMPLATE_SCRIPT_HEADER = """
import os
import sys
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('client')

OUTPUT_FORMAT = '{output_format}'
STDOUT_FILENAME = os.path.expanduser('{stdout_filename}')
STDERR_FILENAME = os.path.expanduser('{stderr_filename}')

if OUTPUT_FORMAT == 'file':
    logger.info('writting output to files stdout={stdout_filename} and stderr={stderr_filename}')
    sys.stdout = open(STDOUT_FILENAME, 'w')
    sys.stderr = open(STDERR_FILENAME, 'w')

"""


def tangle_cells(
    cells, output_format="file", stdout_filename=None, stderr_filename=None
):
    # TODO: eventually support writing output to notebook

    tangled_code = []
    for i, (code, expected_result) in enumerate(cells):
        tangled_code.append('logger.info("beginning execution cell={i}")')
        tangled_code.append(code)
        tangled_code.append('logger.info("completed execution cell={i}")')
    return TEMPLATE_SCRIPT_HEADER + "\n".join(tangled_code)
